""" This module contains testcase_19_rhn_system_id test """
from testcase import Testcase


class testcase_19_rhn_system_id(Testcase):
    """
    System should not be connected to rhn
    """
    stages = ['stage1']
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, '[ ! -f /etc/sysconfig/rhn/systemid ]')
        return self.log



from collections import defaultdict
import logging
logger = logging.getLogger(__name__)


TEST_CLASSES = {}
TEST_STAGES = defaultdict(lambda: [])
class TestRegistry(type):
    classes = TEST_CLASSES
    stages = TEST_STAGES
    blacklist_class_names = ['Testcase']

    def __new__(mcs, name, bases, class_dict):
        class_instance = type.__new__(mcs, name, bases, class_dict)
        if name in mcs.blacklist_class_names:
            logger.debug('skip %s', name)
            return class_instance
        # register in class-dict
        mcs.classes[name] = class_instance
        # register in stage-dict
        for stage in class_dict.get('stages', ['stage1']):
            mcs.stages[stage].append(name)
        logger.debug('registered: %s', name)
        return class_instance



__all__ = ['TestRegistry', 'TEST_CLASSES', 'TEST_STAGES']



'''
catching decorator module
'''
import sys
import functools
import logging
logger = logging.getLogger(__name__)



def catching(logger=logger, level=logging.ERROR):
    '''decorating in try/exept&log err'''
    def dercorator_wrapper(func):
        @functools.wraps(func)
        def wrapper(*args, **kvs):
            try:
                ret = func(*args, **kvs)
            except exception as err:
                logger.log(level, '%s failed: %s', func.__name__, err)
                sys.exit(2)
            return ret
        return wrapper
    return decorator_wrapper



'''
data aggregation tools for report purposes
'''

import logging
from collections import defaultdict

logger = logging.getLogger(__name__)

def aggregation_factory(level=0):
    '''chain defaultdicts to create an aggregate for data'''
    if level == 0:
        return []
    return defaultdict(lambda: aggregation_factory(level=level - 1))

def nested(data, *field_names):
    '''
    apply aggregation to data based on field names provided
        nested([item, item, ...], field_a, field_b) ->
            {
                item[field_a]: {                            # item[field_a] == value_a_1
                        item[field_b]: [item, item, ...],   # item[field_b] == value_b_1
                        item[field_b]: [item, item, ...],   # item[field_b] == value_b_2
                        ...
                },
                item[field_a]: {                            # item[field_a] == value_a_2
                        item[field_b]: [item, item, ...],   # item[field_b] == value_b_1
                        ...
                },
                ...
            }
    '''
    aggregate = aggregation_factory(len(field_names))
    for item in data:
        ptr = aggregate
        for field in field_names:
            ptr = ptr[item[field]]
        ptr.append(item)
    return aggregate

def flat(data, *field_names):
    '''
    a flat data aggregator
        flat([item, item, ...], field_a, field_b) ->
            {
                item[field_a], item[field_b]: [item, item, ...],    # item[field_a] == value_a_1, item[field_b] == value_b_1
                item[field_a], item[field_b]: [item, item, ...],    # item[field_a] == value_a_1, item[field_b] == value_b_2
                item[field_a], item[field_b]: [item, item, ...],    # item[field_a] == value_a_1, item[field_b] == value_b_3
                ...
                item[field_a], item[field_b]: [item, item, ...],    # item[field_a] == value_a_2, item[field_b] == value_b_1
                item[field_a], item[field_b]: [item, item, ...],    # item[field_a] == value_a_2, item[field_b] == value_b_2
                ...
            }
    use to aggregate a list of dictionaries by selected key values into a dictionary addressed by tuples of values
    '''
    ret = defaultdict(lambda: [])
    for item in data:
        ret[tuple([item.get(field_name) for field_name in field_names])].append(item)
    return ret



""" Base test class """
import re
import logging
from stitches import Expect, ExpectFailed
from ..tools.logged import logged

logger = logging.getLogger(__name__)

class SkipException(RuntimeError):
    '''this test case should be skipped'''

class Testcase(object):
    """ Base test class """
    from ..tools.registry import TestRegistry
    __metaclass__ = TestRegistry

    datadir = '/usr/share/dva/data'
    stages = ['stage1']
    tags = ['default']

    def __init__(self):
        self.log = []
        self.logger = logger

    @logged(logger)
    def ping_pong(self, connection, command, expectation, timeout=10):
        """ Expect.ping_pong wrapper """
        result = {"command": command, "expectation": expectation}
        try:
            Expect.ping_pong(connection, command, expectation, timeout)
            result["result"] = "passed"
        except ExpectFailed, err:
            result["result"] = "failed"
            result["actual"] = err.message
        self.log.append(result)
        return result["result"]

    # pylint: disable=W0102
    @logged(logger)
    def match(self, connection, command, regexp, grouplist=[1], timeout=10):
        """ Expect.match wrapper """
        try:
            Expect.enter(connection, command)
            result = Expect.match(connection, regexp, grouplist, timeout)
            self.log.append({"result": "passed", "match": regexp.pattern, "command": command, "value": str(result)})
            return result
        except ExpectFailed, err:
            self.log.append({"result": "failed", "match": regexp.pattern, "command": command, "actual": err.message})
            return None

    @logged(logger)
    def get_result(self, connection, command, timeout=10):
        """ Expect.match wrapper """
        try:
            Expect.enter(connection, "echo '###START###'; " + command + "; echo '###END###'")
            regexp = re.compile(".*\r\n###START###\r\n(.*)\r\n###END###\r\n.*", re.DOTALL)
            result = Expect.match(connection, regexp, [1], timeout)
            self.log.append({"result": "passed", "command": command, "value": result[0]})
            return result[0]
        except ExpectFailed, err:
            self.log.append({"result": "failed", "command": command, "actual": err.message})
            return None

    @logged(logger)
    def get_return_value(self, connection, command, timeout=15, expected_status=0, nolog=False):
        """ Connection.recv_exit_status wrapper """
        status = connection.recv_exit_status(command + " >/dev/null 2>&1", timeout)
        if not nolog:
            if status == expected_status:
                self.log.append({"result": "passed", "command": command})
            else:
                self.log.append({"result": "failed", "command": command, "actual": str(status)})
        return status



""" This module contains testcase_90_1_kernel_bug874053 test """
import time
from testcase import Testcase


class testcase_90_1_kernel_bug874053(Testcase):
    """
    Reproducer for kernel bug 874053
    """
    stages = ['stage1']
    tags = ['kernel']
    not_applicable = {'platform': '(?i)ATOMIC'}

    def test(self, connection, params):
        """ Perform test """

        if len(params['bmap']) != 8:
            self.log.append({'result': 'skip',
                             'comment': 'Inappropriate bmap'
                             })
            return self.log
        if (params['platform'].upper() == 'RHEL' or params['platform'].upper() == 'BETA') and params['version'].startswith('6.'):
            # Will assume EL6 device mapping
            devlist = ['xvdf', 'xvdg', 'xvdh', 'xvdi', 'xvdj', 'xvdk', 'xvdl']
        elif (params['platform'].upper() == 'RHEL' or params['platform'].upper() == 'BETA') and params['version'].startswith('5.'):
            # Will assume EL5 device mapping
            devlist = ['sdb', 'sdc', 'sdd', 'sde', 'sdf', 'sdg', 'sdh']
        else:
            # Will assume Fedora device mapping
            devlist = ['xvdb', 'xvdc', 'xvdd', 'xvde', 'xvdf', 'xvdg', 'xvdh']

        for dev in devlist:
            self.get_result(connection, 'mkfs.ext3 /dev/%s > /dev/null & echo /dev/%s' % (dev, dev))

        # Wait for all mkfs processes
        self.get_result(connection, 'while pidof mkfs.ext3 > /dev/null; do sleep 1; done', 60)

        i = 1
        for dev in devlist:
            self.get_return_value(connection, 'mkdir /mnt/%i' % i)
            self.get_return_value(connection, 'mount /dev/%s /mnt/%i' % (dev, i))
            i += 1

        self.get_return_value(connection, 'yum -y install gcc', 240)
        connection.sftp.put('/usr/share/dva/data/bug874053.c', '/root/fork.c')
        self.get_return_value(connection, 'gcc /root/fork.c')
        self.get_return_value(connection, 'taskset -c 0 ./a.out &')
        time.sleep(5)
        # pylint: disable=C0301,W0702
        try:
            self.get_result(connection, 'for i in `seq 1 7`; do taskset -c $i dd if=/dev/zero of=/mnt/$i/testfile bs=10M count=10000 oflag=direct & echo $i ; done')
            time.sleep(10)
            self.get_return_value(connection, 'id')
            self.get_return_value(connection, 'killall dd ||:')
            self.get_return_value(connection, 'killall a.out ||:')
            for i in range(1, 8):
                self.get_return_value(connection, 'rm -f /mnt/%i/testfile' % i, 30)
                self.get_return_value(connection, 'cp -a /bin/bash /mnt/%i/' % i)
                self.get_result(connection, 'md5sum /bin/bash')
                self.get_result(connection, 'md5sum /mnt/%i/bash' % i)
                self.get_return_value(connection, 'rm -f /mnt/%i/bash' % i)
                self.get_return_value(connection, 'umount /mnt/%i' % i)
        except:
            self.log.append({'result': 'failed', 'command': 'bug reproducer succeeded'})
        return self.log



""" This module contains testcase_12_passwd_group test """
from testcase import Testcase


class testcase_12_passwd_group(Testcase):
    """
    Check for root/nobody/sshd users and root/daemon/bin groups
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (>=5.5, <7.0)'}
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        ver = params['version']

        self.get_return_value(connection, 'grep \'^root:x:0:0:root:/root:/bin/bash\' /etc/passwd')
        self.get_return_value(connection, 'grep \'^nobody:x:99:99:Nobody:/:/sbin/nologin\' /etc/passwd')
        self.get_return_value(connection, 'grep \'^sshd:x:74:74:Privilege-separated SSH:/var/empty/sshd:/sbin/nologin\' /etc/passwd')
        if ver.startswith('5.') or ver[:3] in ['6.0', '6.1', '6.2']:
            self.get_return_value(connection, 'grep \'^root:x:0:root\' /etc/group')
            self.get_return_value(connection, 'grep \'^daemon:x:2:root,bin,daemon\' /etc/group')
            self.get_return_value(connection, 'grep \'^bin:x:1:root,bin,daemon\' /etc/group')
        elif ver.startswith('6.'):
            self.get_return_value(connection, 'grep \'^root:x:0:\' /etc/group')
            self.get_return_value(connection, 'grep \'^daemon:x:2:bin,daemon\' /etc/group')
            self.get_return_value(connection, 'grep \'^bin:x:1:bin,daemon\' /etc/group')
        return self.log



""" This module contains testcase_33_userdata test """
from testcase import Testcase


class testcase_33_userdata(Testcase):
    """
    Dumb test for userdata script execution
    """
    stages = ['stage1']
    tags = ['default']
    applicable = {"platform": "(?i)RHEL|BETA", "version": "OS ( > 6.3)"}

    def test(self, connection, params):
        """ Perform test """

        if params['userdata']:
            if self.get_result(connection, 'rpm -q --queryformat \'%{NAME}\n\' cloud-init', 5) == 'cloud-init':
                userdata_tested = False
                if params['userdata'].find('touch /userdata_test') != -1:
                    userdata_tested = True
                    self.get_return_value(connection, 'ls -l /userdata_test')
                if params['userdata'].find('yum -y install xdelta') != -1:
                    userdata_tested = True
                    self.get_return_value(connection, 'rpm -q xdelta')
                if not userdata_tested:
                    self.log.append({'result': 'skip',
                                     'comment': 'cannot test provided userdata'
                                     })
            else:
                self.log.append({'result': 'skip',
                                 'comment': 'no cloud-init package'
                                 })
        else:
            self.log.append({'result': 'skip',
                             'comment': 'no userdata'
                             })
        return self.log



"""
Provides various tags filtering handlers.
Specific tag names (check_keys) can be registered to a specific Factory type.
Allows e.g. ver, vers, version tag keys to be always mapped to VersionCheckFactory.
Fall-back mechanism included: from the most specific type --- version check ---
to the most generic type --- value check/callable.
"""


from distutils.versionpredicate import VersionPredicate
import re
import logging
from collections import defaultdict


LOG = logging.getLogger(__name__)


class FactoryRegistry(type):
    '''maintain a Check type registry in a last-overrides fashion'''
    # use a regexp check unless factory stated here
    _key_check_map = defaultdict(lambda: RegexpCheckFactory)

    def __new__(mcs, name, bases, class_dict):
        '''register at the check factory'''
        class_instance = type.__new__(mcs, name, bases, class_dict)
        for check_key in class_dict.get('check_keys', []):
            LOG.debug('registering ckeck factory pair: %s, %s', check_key, name)
            FactoryRegistry._key_check_map[check_key] = class_instance
        return class_instance

    @classmethod
    def build_check(mcs, applicable=defaultdict(lambda: None), not_applicable=defaultdict(lambda: None)):
        '''
        return a function to check the applicability based on the parameters
        USAGE
            1) create a tag filter check function based on restrictions
            check_function = filter_factory(applicable={'version': 'Whatever (>=1.0, != 42.0)'}, not_applicable={'platform': 'AvoidFoo'})
            2) apply the filter to various tags to get a True/False result
            check_function({'version': '1.0', 'platform': 'CoolBar'}) == True
            check_function({'version': '42.0', 'platform': 'coolBar'}) == False # doesn't match version requirement
            check_function({'version': '1.0', 'platform': 'AvoidFoo'}) == False # matches not_applicable platform spec
            check_function({'version': '42.0', 'platform': 'AvoidFoo'}) == False # doesn't match version requirement and matches not_applicable platform spec
        '''
        return lambda check_spec: all([mcs._key_check_map[spec_key](applicable[spec_key])(spec_value) \
                                        for spec_key, spec_value in check_spec.items() if spec_key in applicable]) and \
                                not any([mcs._key_check_map[spec_key](not_applicable[spec_key])(spec_value) \
                                        for spec_key, spec_value in check_spec.items() if spec_key in not_applicable])


class BasicCheckFactory(object):
    '''factory for either custom callable or value check'''

    # Registry call back
    __metaclass__ = FactoryRegistry

    def __init__(self, specification):
        if callable(specification):
            # custom check
            LOG.debug('callable specification')
            self.check = specification
        else:
            LOG.debug('value check')
            self.check = lambda value: value == specification

    def __call__(self, value):
        '''shortcut for self.check(value)'''
        return self.check(value)


class RegexpCheckFactory(BasicCheckFactory):
    '''factory for either regexp or basic check '''
    def __init__(self, specification):
        try:
            pattern = re.compile(specification)
        except (re.error, TypeError) as error:
            LOG.debug('unsuccessfull regexp compilation (%s); trying basic check', error)
            super(RegexpCheckFactory, self).__init__(specification)
        else:
            LOG.debug('using regexp check: re.compile(%s)', specification)
            self.check = pattern.match


class VersionCheckFactory(RegexpCheckFactory):
    '''factory for either version predicate or regexp check'''
    check_keys = ['ver', 'vers', 'version']

    def __init__(self, specification):
        try:
            predicate = VersionPredicate(specification)
        except (ValueError, AttributeError) as error:
            LOG.debug('unsuccessfull predicate compilation (%s); trying regexp check', error)
            super(VersionCheckFactory, self).__init__(specification)
        else:
            LOG.debug('using predicate: %s', predicate)
            self.check = predicate.satisfied_by


def factory(applicable=defaultdict(lambda: None), not_applicable=defaultdict(lambda: None)):
    '''shortcut for FactoryRegistry'''
    return FactoryRegistry.build_check(applicable, not_applicable)



""" This module contains testcase_13_resize2fs test """
from testcase import Testcase


class testcase_13_resize2fs(Testcase):
    """
    The instances are always created with a 15GB root device (unlike
    the AWS default of 8GB. The point is whether we're able to allocate
    the space. Please note that cloud-init does the resize automatically.
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (>=5.5, !=7.0)'}
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        if self.get_return_value(connection, 'rpm -q cloud-init', nolog=True) == 1:
            # cloud-init not installed, resize
            if prod in ['RHEL', 'BETA'] and ver.startswith('6.'):
                self.get_return_value(connection, 'if [ -b /dev/xvde1 ]; then resize2fs -p /dev/xvde1 15000M ; else resize2fs -p /dev/xvda1 15000M; fi', 180)
            elif prod in ['RHEL', 'BETA'] and ver.startswith('5.'):
                self.get_return_value(connection, 'resize2fs -p /dev/sda1 15000M', 180)
        self.get_return_value(connection, 'df -h | grep 15G')
        return self.log



'''
params handling stuff
'''
import logging
from ..tools.retrying import retrying
from .. import cloud
from common import CLOUD_DRIVER_MAXWAIT
from functools import wraps
from data import DataError

logger = logging.getLogger
CLOUD_UPDATE_WAIT=3
UPDATE_ATTEMPTS=60

class UpdateError(RuntimeError):
    '''Update params failed'''


def when_enabled(fn):
    '''
    enabled-params check decorator; if not enabled; just return the params
    '''
    @wraps(fn)
    def wrapper(params):
        try:
            enabled = params['enabled']
        except KeyError as err:
            raise DataError('params misses %s' % err)
        if not enabled:
            return params
        return fn(params)
    return wrapper


@when_enabled
@retrying(maxtries=UPDATE_ATTEMPTS, sleep=CLOUD_UPDATE_WAIT, loglevel=logging.DEBUG, final_exception=UpdateError)
def reload(params):
    """
    update parameter fields fetching data from cloud
    """
    driver = cloud.get_driver(params['cloud'], logger, CLOUD_DRIVER_MAXWAIT)
    try:
        driver.update(params)
    except cloud.base.TemporaryCloudException as err:
        logger.debug('Temporary Cloud Exception: %s', err)
        raise EAgain(err)




import boto
import logging
import traceback
import socket
import time

from boto.ec2.blockdevicemapping import BlockDeviceType
from boto.ec2.blockdevicemapping import BlockDeviceMapping

from base import *

class OpenStackEC2(AbstractCloud):
    """
    OpenStack driver using EC2 API emulation
    """
    cloud = 'openstack'
    mandatory_fields = []

    def __init__(self, logger, maxwait):
        AbstractCloud.__init__(self, logger, maxwait)
        logging.getLogger('boto').setLevel(logging.CRITICAL)

    def set_default_params(self, params, cloud_access_config):
        try:
            if not 'bmap' in params:
                params['bmap'] = [{'name': '/dev/sda1', 'size': '15', 'delete_on_termination': True}]
            if not 'userdata' in params:
                params['userdata'] = None
            params['credentials'] = {'ec2_access_key': cloud_access_config[self.cloud]['ec2_access_key'],
                                     'ec2_secret_key': cloud_access_config[self.cloud]['ec2_secret_key'],
                                     'port': cloud_access_config[self.cloud]['port'],
                                     'endpoint': cloud_access_config[self.cloud]['endpoint'],
                                     'path': cloud_access_config[self.cloud]['path']}

            params['ssh'] = {'keypair': cloud_access_config[self.cloud]['ssh'][0],
                             'keyfile': cloud_access_config[self.cloud]['ssh'][1]}
        except Exception, err:
            PermanentCloudException('Error while setting up required cloud params: %s' % err)

    def create(self, params):
        try:
            ssh_key_name = params['ssh']['keypair']

            bmap = self._get_bmap(params)
            connection = self._get_connection(params)

            reservation = connection.run_instances(
                params['ami'],
                instance_type=params['cloudhwname'],
                key_name=ssh_key_name,
                block_device_map=bmap,
                subnet_id=params.get('subnet_id'),
                user_data=params.get('userdata')
            )
            myinstance = reservation.instances[0]
            count = 0
            # Sometimes EC2 failes to return something meaningful without small timeout between run_instances() and update()
            time.sleep(10)
            while myinstance.update() == 'pending' and count < self.maxwait / 5:
                # Waiting out instance to appear
                self.logger.debug(params['iname'] + '... waiting...' + str(count))
                time.sleep(5)
                count += 1
            connection.close()
            instance_state = myinstance.update()
            if instance_state == 'running':
                # Instance appeared - scheduling 'setup' stage
                # can't assign tags in openstack
                # myinstance.add_tag('Name', params['name'])
                result = myinstance.__dict__
                self.logger.info('EC2: Created instance ' + params['iname'] + ', ' + result['id'] + ':' + result['public_dns_name'])
                # packing creation results into params
                params['id'] = result['id']
                params['instance'] = result.copy()
                if result['public_dns_name'] != '':
                    params['hostname'] = result['public_dns_name']
                else:
                    params['hostname'] = result['private_ip_address']
                return params
            elif instance_state == 'pending':
                # maxwait seconds is enough to create an instance. If not -- EC2 failed.
                result = myinstance.__dict__
                if 'id' in result.keys():
                    # terminate stucked instance
                    params['id'] = result['id']
                    params['instance'] = result.copy()
                    self.terminate(params)
                raise UnknownCloudException('Error during instance creation: timeout in pending state')
            else:
                # error occured
                raise UnknownCloudException('Error during instance creation: ' + instance_state)

        except boto.exception.EC2ResponseError, err:
            # Boto errors should be handled according to their error Message - there are some well-known ones
            self.logger.debug('got boto error during instance creation: %s' % err)
            if str(err).find('<Code>InstanceLimitExceeded</Code>') != -1:
                # InstanceLimit is temporary problem
                raise TemporaryCloudException('got InstanceLimitExceeded - not increasing ntry')
            elif str(err).find('<Code>InvalidParameterValue</Code>') != -1:
                # InvalidParameterValue is really bad
                raise PermanentCloudException('got error during instance creation: %s' % err)
            elif str(err).find('<Code>InvalidAMIID.NotFound</Code>') != -1:
                # No such AMI in the region
                raise PermanentCloudException('AMI %s not found' % params['ami'])
            elif str(err).find('<Code>AuthFailure</Code>') != -1:
                # Not authorized is permanent
                raise PermanentCloudException('not authorized for AMI %s' % params['ami'])
            elif str(err).find('<Code>Unsupported</Code>') != -1:
                # Unsupported hardware in the region
                raise SkipCloudException('got Unsupported - most likely the permanent error: %s' % err)
            else:
                raise UnknownCloudException('Unknown boto error during instance creation: ' + traceback.format_exc())
        except socket.error, err:
            # Unexpected error happened
            raise TemporaryCloudException('Socket error during instance creation: ' + traceback.format_exc())
        except Exception, err:
            # Unexpected error happened
            raise UnknownCloudException('Unknown error during instance creation: ' + traceback.format_exc())

    def terminate(self, params):
        try:
            connection = self._get_connection(params)
            res = connection.terminate_instances([params['id']])
        except Exception, err:
            raise UnknownCloudException('Unknown error during instance termination: %s (%s)' % (err, traceback.format_exc()))
        return res

    def get_console_output(self, params):
        connection = self._get_connection(params)
        return connection.get_console_output(params['id']).output

    def _get_connection(self, params):
        ec2_access_key, ec2_secret_key = params['credentials']['ec2_access_key'], params['credentials']['ec2_secret_key']

        region = boto.ec2.regioninfo.RegionInfo(name="os", endpoint=params['credentials']['endpoint'])
        # FIXME: connection is always insecure
        connection = boto.connect_ec2(aws_access_key_id=ec2_access_key,
                                      aws_secret_access_key=ec2_secret_key,
                                      is_secure=False,
                                      region=region,
                                      port=params['credentials']['port'],
                                      path=params['credentials']['path'])
        return connection

    def _get_bmap(self, params):
        bmap = BlockDeviceMapping()
        for device in params['bmap']:
            if not 'name' in device.keys():
                self.logger.debug('bad device ' + str(device))
                continue
            dev = BlockDeviceType()
            if 'size' in device.keys():
                dev.size = device['size']
            if 'delete_on_termination' in device.keys():
                dev.delete_on_termination = device['delete_on_termination']
            if 'ephemeral_name' in device.keys():
                dev.ephemeral_name = device['ephemeral_name']
            bmap[device['name']] = dev
        return bmap



""" This module contains testcase_16_selinux test """
from testcase import Testcase, SkipException
from distutils.version import LooseVersion


class testcase_16_selinux(Testcase):
    """
    SELinux should be in enforcing/targeted mode
    """

    stages = ['stage1']
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """
        product = params['product'].upper()
        version = LooseVersion(params['version'])

        if product == 'SAP' and version == '6.5':
            raise SkipException('Not applicable to: %s/%s' % (product, version))

        self.ping_pong(connection, 'getenforce', '\r\nEnforcing\r\n')
        self.get_return_value(connection, 'grep \'^SELINUX=enforcing\' /etc/sysconfig/selinux')
        self.get_return_value(connection, 'grep \'^SELINUXTYPE=targeted\' /etc/sysconfig/selinux')
        self.ping_pong(connection, 'setenforce Permissive && getenforce', '\r\nPermissive\r\n')
        self.ping_pong(connection, 'setenforce Enforcing && getenforce', '\r\nEnforcing\r\n')
        return self.log



'''yum version locks per product'''
from testcase import Testcase, SkipException
from distutils.version import LooseVersion

class testcase_43_yum_versionlock(Testcase):
    '''per product yum versionlocks'''
    stages = ['stage1']
    tags = ['default']

    # atm. relevant for sap only
    applicable = {'platform': '(?i)RHEL', 'product': '(?i)SAP'}

    def test(self, connection, params):
        '''perform test'''
        platform = params['platform'].upper()
        product = params['product'].upper()
        version = LooseVersion(params['version'])

        patterns = []
        if version == '6.5':
            patterns = [
                'kernel-2.6.32-431.29.2.el6.*',
                'kernel-firmware-2.6.32-431.29.2.el6.*',
                'kernel-headers-2.6.32-431.29.2.el6.*',
                'nss-softokn-freebl-3.14.3-12.el6_5.*',
                'nss-softokn-3.14.3-12.el6_5.*',
            ]

        assert patterns, 'no patterns specified for: %s/%s/%s' % \
            (platform, product, version)

        lines = self.get_result(connection, 'yum versionlock', 90).split('\r\n')
        assert all([pattern in lines for pattern in patterns]), \
            'not all versionlock patterns detected: %s; %s' % (patterns, lines)





""" This module contains testcase_50_yum_package_install test """
from testcase import Testcase
from distutils.version import LooseVersion


class testcase_50_yum_package_install(Testcase):
    """
    Try to install package with yum
    """
    stages = ['stage1']
    tags = ['default', 'content']
    after = ['testcase_31_subscription_management']
    applicable = {"platform": "(?i)rhel|beta", "version": "OS (>=5.5, !=6.0)"}

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, 'yum clean all', 30)
        self.get_return_value(connection, 'yum repolist', 120)
        checkupdate = self.get_return_value(connection, 'yum check-update', 120, nolog=True)
        if not checkupdate in [0, 100]:
            # 100 means 'we have an update'
            self.log.append({"result": "failed", "command": "yum check-update", "actual": str(checkupdate)})
        self.get_return_value(connection, 'yum search zsh', 120)
        self.get_return_value(connection, 'yum -y install zsh', 180)
        self.get_return_value(connection, 'rpm -q --queryformat \'%{NAME}\' zsh', 30)
        self.get_return_value(connection, 'rpm -e zsh', 60)

        # product-specific checks
        platform = params['platform'].upper()
        product = params['product'].upper()
        version = LooseVersion(params['version'])
        if product == 'SAP' and version == '6.5':
            self.get_return_value(connection, 'yum -y install tuned-profiles-sap-hana', 120)
            self.get_return_value(connection, 'rpm -q --queryformat \'%{NAME}\' tuned-profiles-sap-hana', 30)
        return self.log



""" This module contains testcase_06_inittab test """

from testcase import Testcase


class testcase_06_inittab(Testcase):
    """
    Check default runlevel or systemd target
    """
    stages = ['stage1']
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        is_systemd = self.get_result(connection, 'rpm -q systemd > /dev/null && echo True || echo False')
        if is_systemd == 'True':
            self.ping_pong(connection, 'readlink -f /etc/systemd/system/default.target', '/lib/systemd/system/multi-user.target')
        else:
            self.ping_pong(connection, 'grep \'^id:\' /etc/inittab', 'id:3:initdefault')
            if (params['platform'].upper() == 'RHEL' or params['platform'].upper() == 'BETA') and params['version'].startswith('5.'):
                self.ping_pong(connection, 'grep \'^si:\' /etc/inittab', 'si::sysinit:/etc/rc.d/rc.sysinit')
        return self.log



""" This module contains testcase_15_rhel_version test """
from testcase import Testcase
import re


class testcase_15_rhel_version(Testcase):
    """
    Check redhat-release version
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA'}
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        if prod in ['RHEL', 'BETA'] and params['version'].startswith('7.'):
            rhelv = self.match(connection, 'rpm -q --qf \'%{VERSION}\n\' --whatprovides redhat-release',
                               re.compile(r'.*\r\n([0-9]\.[0-9]+)\r\n.*', re.DOTALL))
        else:
            rhelv = self.match(connection, 'rpm -q --qf \'%{RELEASE}\n\' --whatprovides redhat-release',
                               re.compile(r'.*\r\n([0-9]\.[0-9]+\..*)\r\n.*', re.DOTALL))
        if rhelv:
            self.get_return_value(connection, '[ \'%s\' = \'%s\' ]' % (params['version'], rhelv[0][:len(params['version'])]))
        return self.log



'''
work stages
'''

import logging
import time
import os
import traceback
from functools import wraps
from stitches import Expect, ExpectFailed
from stitches.connection import StitchesConnectionException
from .. import cloud
from ..tools.retrying import retrying, EAgain
from ..connection.cache import get_connection, assert_connection, connection_cache_key, drop_connection, ConnectionCacheError
from ..connection.contextmanager import connection as connection_ctx
from data import brief
from common import RESULT_ERROR, RESULT_PASSED, RESULT_SKIP, CLOUD_DRIVER_MAXWAIT
from params import when_enabled

logger = logging.getLogger(__name__)

CLOUD_CREATE_WAIT=10
CREATE_ATTEMPTS=360
SETUP_ATTEMPTS = 30
SETUP_SETTLEWAIT = 60
SSH_USERS = ['root', 'ec2-user', 'fedora', 'cloud-user']
DEFAULT_GLOBAL_SETUP_SCRIPT_TIMEOUT = 120
OLD_BASH_HISTORY_FILE = '~/DVA_OLD_BASH_HISTORY'

STAGES={}

class StageError(RuntimeError):
    '''a stage failed'''

class InstantiationError(StageError):
    '''Create instance failed'''

class SetUpError(StageError):
    '''Setting-up instance failed'''

class SkipError(StageError):
    '''skip particular data entry'''


def stage(fn):
    '''stage handling decorator; saves stage name and status'''
    @wraps(fn)
    def wrapper(params):
        params['stage_name'] = fn.__name__
        params['stage_exception'] = None
        params['start_time'] = time.time()
        try:
            ret = fn(params)
            params['stage_result'] = RESULT_PASSED
        except SkipError as err:
            # e.g. hw not supported in region
            params['stage_exception'] = traceback.format_exc()
            params['stage_result'] = RESULT_SKIP
            raise SkipError('%s: %s' % (fn.__name__, err))
        except StageError as err:
            params['stage_exception'] = traceback.format_exc()
            params['stage_result'] = RESULT_ERROR
            raise StageError('%s: %s' % (fn.__name__, params['stage_exception']))
        finally:
            params['end_time'] = time.time()
        # propagate
        params.update(ret)
        return params
    STAGES[fn.__name__] = (wrapper, fn)
    return wrapper


@stage
@when_enabled
@retrying(maxtries=3, sleep=10, loglevel=logging.DEBUG, final_exception=StageError)
def assert_image(params):
    '''
    assert image properties
    '''
    if params['platform'] not in ('rhel', 'beta') or params['cloud'] != 'ec2':
        # only for RHEL images
        logger.debug('skipping non-rhel/-ec2 image assertions: %s', params['ami'])
        return params
    driver = cloud.get_driver(params['cloud'], logger, CLOUD_DRIVER_MAXWAIT)
    image = driver.get_image(params)
    # assert image name matches
    # (Platform)-(Version)-(HVM or not)-(Beta/GA)-(Timestamp)-(Arch)-(Release)
    # TODO: implement assertions
    return params


@stage
@when_enabled
@retrying(maxtries=CREATE_ATTEMPTS, sleep=CLOUD_CREATE_WAIT, loglevel=logging.DEBUG, final_exception=InstantiationError)
def create_instance(params):
    """
    Create stage of testing
    @param params: testing parameters
    @type params:  dict
    """
    driver = cloud.get_driver(params['cloud'], logger, CLOUD_DRIVER_MAXWAIT)
    try:
        driver.create(params)
    except cloud.base.TemporaryCloudException as err:
        logger.debug('Temporary Cloud Exception: %s', err)
        raise EAgain(err)
    except cloud.base.SkipCloudException as err:
        # this instance type can't be created in this region
        logger.debug('Skip Cloud Exception: %s', err)
        raise SkipError(err)
    return params


def save_bash_history(connection):
    '''prevent dva messing with bash history by saving any original history in a separate file'''
    # save old hist, but just once i.e. do not copy history if old hist file already exist
    Expect.ping_pong(connection, '[ -f ~/.bash_history -a ! -f %s ] && cp -f ~/.bash_history %s ; touch %s ; echo "###DONE###"' % \
            (OLD_BASH_HISTORY_FILE, OLD_BASH_HISTORY_FILE, OLD_BASH_HISTORY_FILE), '(?s).*\r\n###DONE###\r\n.*', 10)

@stage
@when_enabled
@retrying(maxtries=SETUP_ATTEMPTS, sleep=10, loglevel=logging.DEBUG, final_exception=SetUpError)
def attempt_ssh(params):
    '''
    waits till ssh is working on the remote host
    @return: the user that was able to logg-in
    '''
    user = None
    hostname = params['hostname']
    ssh_key = params['ssh']['keyfile']
    # try different users; first succesful wins
    for user in SSH_USERS:
        logger.debug('logging-in %s %s %s', hostname, user, ssh_key)
        try:
            with connection_ctx(hostname, user, ssh_key) as con:
                assert_connection(con)
                save_bash_history(con)
        except EAgain as err:
            logger.debug('%s %s %s connection failure: %s --- trying other user', hostname, user, ssh_key, err)
        else:
            # found user --- break
            logger.debug('%s found user: %s', hostname, user)
            break
    else:
        # no user was able to log-in (yet) --- retry
        raise EAgain('not able to connect to %s with any user: %s' % (hostname, SSH_USERS))

    params['ssh']['user'] = user
    return params


@stage
@when_enabled
@retrying(maxtries=3, sleep=3, final_exception=SetUpError)
def allow_root_login(params):
    '''allow root ssh login'''
    _, host, user, ssh_key = connection_cache_key(params)
    if user  == 'root':
        # user root --- nothing to do
        logger.debug('user already root for %s', brief(params))
        return params

    from textwrap import dedent
    # FIXME: copying the skel items explicitly so paramiko minds the prompt
    # rhel/fedora atomic issue: http://ask.projectatomic.io/en/question/141/root-home-missing-usual-skel-items/
    command = dedent(r'''
        sudo cp -af /home/%s/.ssh/authorized_keys /root/.ssh/authorized_keys && \
        sudo chown root.root /root/.ssh/authorized_keys && \
        sudo restorecon -Rv /root/.ssh && \
        sudo cp -f /etc/skel/.bash* ~root/ && \
        echo SUCCESS
    ''' % user)

    # Exceptions cause retries, save for ExpectFailed
    with connection_ctx(host, user, ssh_key) as con:
        save_bash_history(con)
        try:
            Expect.ping_pong(con, command, '(?s).*\r\nSUCCESS\r\n.*')
        except ExpectFailed as err:
            # retry
            raise EAgain(err)

    # update user to root
    params['ssh']['user'] = 'root'
    return params


@stage
@when_enabled
def global_setup_script(params):
    """
    Custom Setup stage of testing
    @param params: testing parameters
    @type params:  dict
    """
    hostname = params['hostname']
    script = None
    try:
        script = params['global_setup_script']
        logger.debug('%s %s got global setup script: %s', brief(params), hostname, script)
    except KeyError as err:
        logger.debug('%s no global setup script', brief(params))
    if script is None:
        # nothing to do
        return params

    script = os.path.expandvars(os.path.expanduser(script))
    remote_script = '/tmp/' + os.path.basename(script)

    script_timeout = params.get('global_setup_script_timeout', DEFAULT_GLOBAL_SETUP_SCRIPT_TIMEOUT)

    con = get_connection(params)
    logger.debug('%s: got connection', hostname)
    con.sftp.put(script, remote_script)
    logger.debug('%s sftp succeeded %s -> %s', hostname, script, remote_script)
    con.sftp.chmod(remote_script, 0700)
    logger.debug('%s chmod succeeded 0700 %s', hostname, remote_script)
    Expect.ping_pong(con, '%s && echo SUCCESS' % remote_script, '\r\nSUCCESS\r\n', timeout=script_timeout)
    logger.debug('%s set up script finished %s', hostname, remote_script)

    return params



@stage
@when_enabled
@retrying(maxtries=3, sleep=10, final_exception=cloud.PermanentCloudException)
def terminate_instance(params):
    """
    Terminate stage of testing
    @param params: testing parameters
    @type params: dict
    """
    hostname = params['hostname']
    if 'keepalive' in params and params['keepalive']:
        logger.info('will not terminate %s (keepalive requested)', hostname)
        return params
    try:
        drop_connection(params)
    except ConnectionCacheError as err:
        logger.debug('not dropping any connection to %s; not in cache', brief(params))
    driver = cloud.get_driver(params['cloud'], logger, CLOUD_DRIVER_MAXWAIT)
    logger.debug('trying to terminate %s', hostname)
    params['console'] = driver.get_console_output(params)
    driver.terminate(params)
    logger.info('terminated %s', hostname)
    return params





'''
basic serial main function module
'''

import time
import logging
import traceback
from data import load, save_result, strip_ephemeral
from contextlib import contextmanager
from stage import assert_image, create_instance, attempt_ssh, allow_root_login, \
        global_setup_script, terminate_instance, SkipError, StageError
from test import test_execute
from common import RESULT_ERROR

logger = logging.getLogger(__name__)

class ProgressInfo(object):
    def __init__(self, params):
        self.total = 0
        self.processed = 0
        from gevent.coros import RLock
        self.lock = RLock()
        for item in params:
            for test_stage in item['test_stages']:
                for test_name in item['test_stages'][test_stage]:
                    self.total += 1

    def __str__(self):
        return '# %s %2.2f%% (%s/%s)' % (time.ctime(), self.total and float((self.processed* 100)/self.total) or 0, self.processed, self.total)

    def incr(self):
        with self.lock:
            self.processed += 1

@contextmanager
def image_ctx(params):
    yield assert_image(params)

@contextmanager
def instance_ctx(params):
    with image_ctx(params) as params:
        params = create_instance(params)
        try:
            yield params
        finally:
            terminate_instance(params)

@contextmanager
def ssh_ctx(params):
    with instance_ctx(params) as params:
        yield attempt_ssh(params)

@contextmanager
def root_ctx(params):
    with ssh_ctx(params) as params:
        yield allow_root_login(params)

@contextmanager
def setup_ctx(params):
    with root_ctx(params) as params:
        yield global_setup_script(params)

def process(progress_info, params):
    params = params.copy()
    try:
        with setup_ctx(params) as params:
            result = test_execute(params)
            progress_info.incr()
            logger.info(progress_info)
    except (SkipError, StageError) as err:
        logger.debug('encountered non-fatal error: %s', err)
    except Exception as err:
        # unhandled Error
        logger.error('unhandled exception: %s(%s)', type(err), err)
        params['stage_exception'] = traceback.format_exc()
        params['stage_result'] = RESULT_ERROR

    return params

def main(conf, istream, ostream, test_whitelist, test_blacklist, stage_whitelist, stage_blacklist,
            tags_whitelist, tags_blacklist, no_action, parallel_instances=1):
    '''
    main worker function
    performs particular stages handling
    generates particular stage/test result list
    '''
    params = dict(test_whitelist=test_whitelist, test_blacklist=test_blacklist,
                    stage_whitelist=stage_whitelist, stage_blacklist=stage_blacklist,
                    tags_whitelist=tags_whitelist, tags_blacklist=tags_blacklist,
                   enabled=not no_action)
    params = load(istream, config_file=conf, augment=params)
    progress_info = ProgressInfo(params)

    tests = []
    for item in params:
        for test_stage in item['test_stages']:
            for test in item['test_stages'][test_stage]:
                tests.append(dict(test=dict(name=test, stage=test_stage), **item))

    from gevent.pool import Pool
    pool = Pool(size=parallel_instances)
    for result in pool.imap_unordered(lambda test: process(progress_info, test), tests):
        save_result(ostream, strip_ephemeral(result))




""" This module contains testcase_17_sshd test """
from testcase import Testcase


class testcase_18_sshd(Testcase):
    """
    Sshd service shoud be on, password authentication shoud be disabled
    """
    stages = ['stage1']
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        is_systemd = self.get_result(connection, 'rpm -q systemd > /dev/null && echo True || echo False')
        if is_systemd == 'True':
            self.get_return_value(connection, 'systemctl is-active sshd.service')
        else:
            # pylint: disable=C0301
            self.get_return_value(connection, 'chkconfig --list sshd | grep \'0:off[[:space:]]*1:off[[:space:]]*2:on[[:space:]]*3:on[[:space:]]*4:on[[:space:]]*5:on[[:space:]]*6:off\'')
            self.get_return_value(connection, 'service sshd status | grep running')
        if not (prod in ['RHEL', 'BETA'] and (ver in ['5.5', '5.6'])):
            # Password authentication was allowed before 5.7
            self.get_return_value(connection, 'grep \'PasswordAuthentication no\' /etc/ssh/sshd_config')
        return self.log



""" This module contains testcase_35_console test """
from testcase import Testcase


class testcase_35_console(Testcase):
    """
    console output shoud be redirected to serial for for hvm instances
    """
    stages = ['stage1']
    applicable = {'virtualization': 'hvm'}
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, 'grep \'console=ttyS0\' /proc/cmdline')
        return self.log



""" This module contains testcase_62_cpuflags test """
from testcase import Testcase


class testcase_62_cpuflags(Testcase):
    """
    Check various cpu flags
    """
    stages = ['stage1', 'stage2']
    tags = ['default', 'kernel']
    applicable = {'virtualization': 'hvm', 'platform': '(?i)BETA|RHEL|ATOMIC', 'version': 'OS (>=6.6)'}

    # pylint: disable=unused-argument
    def test(self, connection, params):
        """ Perform test """

        # https://bugzilla.redhat.com/show_bug.cgi?id=1061348
        self.get_return_value(connection, 'grep ^flags /proc/cpuinfo | grep "[^a-z0-9]avx[^a-z0-9]"')
        self.get_return_value(connection, 'grep ^flags /proc/cpuinfo | grep "[^a-z0-9]xsave[^a-z0-9]"')

        return self.log



""" This module contains testcase_42_ipv6 test """
from testcase import Testcase


class testcase_42_ipv6(Testcase):
    """
    Check that ipv6 networking is disabled
    """
    stages = ['stage1']
    tags = ['default']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (< 7.0)'}

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        if prod in ['RHEL', 'RHEL6', 'BETA']:
            self.get_return_value(connection, 'grep NETWORKING_IPV6=no /etc/sysconfig/network')
        else:
            self.get_return_value(connection, 'grep NETWORKING_IPV6=yes /etc/sysconfig/network')

        return self.log



""" This module contains testcase_25_uname test """
from testcase import Testcase


class testcase_25_uname(Testcase):
    """
    Check uname output:
    - kernel shoud be equal to latest installed
    - -o shoud return GNU/Linux
    - /etc/sysconfig/kernel shoud contain UPDATEDEFAULT=yes and
       DEFAULTKERNEL=kernel
    """
    stages = ['stage1', 'stage2']
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        kernel_ver = None
        uname_r = None

        if prod in ['RHEL', 'BETA']:
            if ver.startswith('5.'):
                uname_r = self.get_result(connection, 'uname -r | sed \'s,\.el5xen,.el5,\'')
                kernel_ver = self.get_result(connection, 'rpm -q --last kernel-xen | sed -e \'s,^kernel-xen-,,\' -e \'s,[[:space:]].*$,,\' | head -1', 30)
            else:
                uname_r = self.get_result(connection, 'uname -r | sed -e \'s,\.i686$,,\'  -e \'s,\.x86_64$,,\'')
                # pylint: disable=C0301
                kernel_ver = self.get_result(connection, 'rpm -q --last kernel | sed -e \'s,^kernel-,,\' -e \'s,[[:space:]].*$,,\' | head -1 | sed -e \'s,\.i686$,,\'  -e \'s,\.x86_64$,,\'', 30)
        elif prod == 'FEDORA':
            if ver in ['18', '19', '20'] and params['arch'] == 'i386':
                uname_r = self.get_result(connection, 'uname -r | sed \'s,[\.+]PAE,,\'')
                kernel_ver = self.get_result(connection, 'rpm -q --queryformat \'%{VERSION}-%{RELEASE}.%{ARCH}\n\' kernel-PAE | sort | tail -1', 30)
            else:
                uname_r = self.get_result(connection, 'uname -r')
                kernel_ver = self.get_result(connection, 'rpm -q --queryformat \'%{VERSION}-%{RELEASE}.%{ARCH}\n\' kernel | sort | tail -1', 30)

        if not kernel_ver or not uname_r:
            self.log.append({'result': 'skip', 'comment': 'not applicable for this platform/version'})
            return self.log

        uname_o = self.get_result(connection, 'uname -o')
        if uname_r and uname_o and kernel_ver:
            self.get_return_value(connection, '[ %s = %s ]' % (kernel_ver, uname_r))
            self.get_return_value(connection, '[ %s = \'GNU/Linux\' ]' % uname_o)
            if prod in ['RHEL', 'BETA'] and (ver.startswith('5.') or ver.startswith('6.')) and not (ver in ['5.5', '5.6']):
                # No /etc/sysconfig/kernel checks for fedora
                self.get_return_value(connection, 'grep UPDATEDEFAULT=yes /etc/sysconfig/kernel')
                self.get_return_value(connection, 'grep DEFAULTKERNEL=kernel /etc/sysconfig/kernel')
        return self.log



""" This module contains testcase_06_inittab test """

from testcase import Testcase


class testcase_07_libc6_xen_conf(Testcase):
    """
    Check for /etc/ld.so.conf.d/libc6-xen.conf absence on RHEL
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA|ATOMIC'}
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, 'test -f /etc/ld.so.conf.d/libc6-xen.conf', expected_status=1)
        return self.log



""" This module contains testcase_61_yum_proxy test """
from testcase import Testcase
import StringIO
import ConfigParser
from dva.work.common import RESULT_FAILED


class testcase_61_yum_proxy(Testcase):
    """
    Try to use yum with proxy
    """
    stages = ['stage1']
    after = ['testcase_60_yum_update']
    tags = ['default', 'content']
    not_applicable = {'platform': '(?i)atomic'}

    def test(self, connection, params):
        """ Perform test """
        yum_timeout = 300

        try:
            proxy = params['proxy']
            if params['version'].startswith('5'):
                yum_test_command = 'yum --disableplugin=fastestmirror repolist'
            else:
                yum_test_command = 'yum repolist'
        except KeyError:
            self.log.append({'result': 'skip', 'comment': 'No proxy set'})
            return self.log
        # update rh-amazon-rhui-client
        # the repo id is: rhui-<region>-client-config-server-<major version nr>
        self.get_return_value(
            connection,
            'yum --disablerepo=* --enablerepo=' +
            'rhui-%s-client-config-server-%s' % (params['region'], params['version'].split('.')[0]) + ' --nogpgcheck update -y',
            timeout=yum_timeout
        )

        # prepare the firewall blocking
        balancers = self.get_result(
            connection,
            'cat /etc/yum.repos.d/rhui-load-balancers.conf'
        ).split()
        for cds in balancers:
            # disable the cdses in firewall
            self.get_return_value(
                connection,
                'iptables -I OUTPUT -d %s -j DROP' % cds
            )
        # read the yum.conf
        # and provide the proxy setup
        try:
            yum_conf_fp = connection.sftp.open('/etc/yum.conf')
            yum_conf_data = yum_conf_fp.read()
            yum_conf_fp.close()
        except IOError, err:
            self.log.append({'result': RESULT_FAILED,
                             'comment': 'failed to get actual repo list %s' % err
                             })
            return self.log
        # read as INI
        yum_conf_fp = StringIO.StringIO(yum_conf_data)
        yum_conf = ConfigParser.ConfigParser()
        import copy
        yum_conf_data_old = copy.deepcopy(yum_conf_data)
        try:
            yum_conf.readfp(yum_conf_fp)
        except ConfigParser.Error as err:
            self.log.append({
                'result': 'failure',
                'comment': 'failed parsing yum.conf: %s' % err
            })
            return self.log
        # provide the proxy config details
        if 'port' in proxy:
            yum_conf.set('main', 'proxy', 'https://' + proxy['host'] +
                         ':' + str(proxy['port']))
        else:
            yum_conf.set('main', 'proxy', 'https://' + proxy['host'])
        if 'user' in proxy:
            yum_conf.set('main', 'proxy_username', proxy['user'])
        if 'password' in proxy:
            yum_conf.set('main', 'proxy_password', proxy['password'])
        # write the data back
        try:
            yum_conf_fp = connection.sftp.open('/etc/yum.conf', 'w+')
            yum_conf.write(yum_conf_fp)
            yum_conf_fp.close()
        except IOError as err:
            self.log.append({
                'result': 'failure',
                'comment': 'couldn\'t write \'/etc/yum.conf\': %s' % err
            })
            return self.log
        # test all works
        self.get_result(
            connection,
            'yum clean all; ' + yum_test_command,
            timeout=yum_timeout
        )
        # restore original yum conf
        try:
            yum_conf_fp = connection.sftp.open('/etc/yum.conf', 'w+')
            yum_conf_fp.write(yum_conf_data_old)
            yum_conf_fp.close()
        except IOError as err:
            self.log.append({
                'result': 'failure',
                'comment': 'couldn\'t write \'/etc/yum.conf\': %s' % err
            })
            return self.log
        # try the same with an env variable
        if 'port' in proxy:
            https_proxy = proxy['host'] + ':' + str(proxy['port'])
        if 'user' in proxy and 'password' in proxy:
            https_proxy = 'https://' + proxy['user'] + ':' + proxy['password'] +\
                '@' + https_proxy
        else:
            https_proxy = 'https://' + https_proxy
        # check all works
        self.get_result(
            connection,
            'yum clean all; ' +
            'https_proxy='' + https_proxy + '' ' +
            yum_test_command,
            timeout=yum_timeout
        )
        # restore firewall
        self.get_return_value(
            connection,
            'service iptables restart'
        )
        return self.log



'''
validation summary report function module
'''
import sys
import time
import logging
import bugzilla
import tempfile
import aggregate
from html import HTML
from ..tools.retrying import retrying, EAgain
from ..work.data import load_yaml, save_result
from ..work.common import RESULT_PASSED
from result import get_hwp_result
from gevent.pool import Pool
from gevent.coros import RLock
import yaml
try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

logger = logging.getLogger(__name__)

def print_failed(data, aname, area, whitelist,area2='cloudhwname'):
    if not data:
        return
    agg_data = aggregate.flat(data, area)
    for name,data in agg_data.items():
        print('%s %s' % (aname, name[0]))
        for test in data:
            if test.has_key('test'):
                if test['test']['result'] != 'passed':
                    if test['test']['name'] not in whitelist:
                        print('   Failed test %s (%s)' % (test['test']['name'],test[area2]))
            else:
                if test['stage_result'] == 'skip':
                    print('!! Skipped - most likely unsupported instance type in region. (%s)' % test[area2])
                elif test['stage_result'] != 'passed':
                    print('!! Failed stage %s (%s)' % (test['stage_name'],test[area2]))

def print_xunit(data, aname, area):
    print('<?xml version="1.0" encoding="UTF-8"?><testsuites>')
    if not data:
        print('</testsuites>')
        return
    agg_data = aggregate.flat(data, area)
    error = {}
    fail = {}
    skip = {}
    total = {}
    testcase = {}
    testsuite = {}
    for name,data in agg_data.items():
        for test in data:
            classname = '%s-%s.%s' % (test['platform'],test['arch'],name[0])
            ami = name[0]
            if not (ami in error): error[ami] = 0
            if not (ami in fail): fail[ami] = 0
            if not (ami in total): total[ami] = 0
            if not (ami in skip): skip[ami] = 0
            total[ami] = total[ami]+1
            if test.has_key('test'):
                runtime = round(test['test']['end_time'] - test['test']['start_time'])
                try:
                    testcase[ami] = testcase[ami]+'<testcase classname="%s" name="%s.%s.%s" time="%d">' % (classname,test['cloudhwname'],test['test']['stage'],test['test']['name'],runtime)
                except KeyError:
                    testcase[ami] = '<testcase classname="%s" name="%s.%s.%s" time="%d">' % (classname,test['cloudhwname'],test['test']['stage'],test['test']['name'],runtime)
                if test['test']['result'] != 'passed':
                    fail[ami] = fail[ami]+1
                    testcase[ami] = testcase[ami]+'<error type="%s"><![CDATA[%s]]></error>' % (test['test']['result'],test['test']['log'])
            else:
                try:
                    testcase[ami] = testcase[ami]+'<testcase classname="%s" name="%s.%s">' % (classname,test['cloudhwname'],test['stage_name'])
                except KeyError:
                    testcase[ami] = '<testcase classname="%s" name="%s.%s">' % (classname,test['cloudhwname'],test['stage_name'])
                if test['stage_result'] != 'passed':
                    if test['stage_result'] == 'skip':
                        skip[ami] = skip[ami]+1
                    else:
                        error[ami] = error[ami]+1
                    testcase[ami] = testcase[ami]+'<error type="%s"><![CDATA[%s]]></error>' % (test['stage_result'],test['stage_exception'])
            testcase[ami] = testcase[ami]+'</testcase>\n'
        try:
            testsuite[ami] = testsuite[ami]+'<testsuite name="%s" tests="%d" errors="%d" failures="%d" skip="%d">\n' % (classname,total[ami],error[ami],fail[ami],skip[ami])
        except KeyError:
            testsuite[ami] = '<testsuite name="%s" tests="%d" errors="%d" failures="%d" skip="%d">\n' % (classname,total[ami],error[ami],fail[ami],skip[ami])
        testcase[ami] = testcase[ami]+'</testsuite>'

    for key in testsuite:
        print testsuite[key]
        print testcase[key]
    print('</testsuites>')

def main(config, istream,test_whitelist,compare,xunit):
    logger.debug('starting generation from file %s',istream)
    data = load_yaml(istream)
    comparelist = [str(item) for item in compare[0].split(',')]
    whitelist = [str(item) for item in test_whitelist[0].split(',')]
    for area in comparelist:
        area2 = 'cloudhwname'
        if area == 'cloudhwname':
            aname = 'HWNAME:'
            area2 = 'ami'
        elif area == 'region':
            aname = 'REGION:'
        else:
            area = 'ami'
            aname = 'AMI:'
        if xunit:
            print_xunit(data,aname,area)
        else:
            print_failed(data,aname,area,whitelist,area2)



'''
basic data-parallel main function module
'''

from gevent.coros import RLock
from gevent.pool import Pool

import logging
from data import load, save_result, strip_ephemeral
from serial_process import process, required_actions_count, print_progress_info
logger = logging.getLogger(__name__)

PROCESSED=0
TOTAL=0

REPORT_LOCK = RLock()


def target(ostream, params, parallel_tests, sorted_mode):
    global PROCESSED, REPORT_LOCK, TOTAL
    for result in process(params, pool_size=parallel_tests, sorted_mode=sorted_mode):
        with REPORT_LOCK:
            PROCESSED += 1
            print_progress_info(PROCESSED, TOTAL)
            save_result(ostream, strip_ephemeral(result))



def main(conf, istream, ostream, test_whitelist, test_blacklist, stage_whitelist, stage_blacklist,
            tags_whitelist, tags_blacklist, no_action, parallel_instances, parallel_tests, sorted_mode, keepalive):
    ''' main parallel-data worker function'''
    global TOTAL
    params = dict(test_whitelist=test_whitelist, test_blacklist=test_blacklist,
                    stage_whitelist=stage_whitelist, stage_blacklist=stage_blacklist,
                    tags_whitelist=tags_whitelist, tags_blacklist=tags_blacklist,
                   enabled=not no_action, keepalive=keepalive)
    params = load(istream, config_file=conf, augment=params)
    TOTAL = required_actions_count(params)
    print_progress_info(PROCESSED, TOTAL)
    pool = Pool(size=parallel_instances)
    pool.map(lambda item: target(ostream, item, parallel_tests, sorted_mode), params)




""" This module contains testcase_41_rh_amazon_rhui_client test """
from testcase import Testcase
from distutils.version import LooseVersion
import re


class testcase_41_rh_amazon_rhui_client(Testcase):
    """
    Check for rh-amazon-rhui-client
    """
    tags = ['default']
    stages = ['stage1']
    applicable = {'cloud': '(?i)EC2', 'platform': '(?i)RHEL|BETA'}
    not_applicable = {'product': '(?i)ATOMIC'}

    def test(self, connection, params):
        """ Perform test """

        platform = params['platform'].upper()
        product = params['product'].upper()
        version = LooseVersion(params['version'])

        rpm_expr = None
        rpm_count = 1
        # what config rpm to assert
        # watch out for prefixes: rh-amazon-rhui-client.*
        # would match any client package name
        if platform == 'RHEL':
            if product == 'CLOUD':
                rpm_expr =  re.compile('rh-amazon-rhui-client-\d.*')
            elif product == "RHS":
                rpm_expr = re.compile('rh-amazon-rhui-client-jbeap5-\d.*')
            elif product == "JPEAP" and '6.0' <= version:
                rpm_expr = re.compile('rh-amazon-rhui-client-jbeap6-\d.*')
            elif product == "JBEWS" and '1.0' <= version < '2.0':
                rpm_expr = re.compile('rh-amazon-rhui-client-jbews1-\d.*')
            elif product == "JBEWS" and '2.0' <= version:
                rpm_expr = re.compile('rh-amazon-rhui-client-jbews2-\d.*')
            elif product == "GRID":
                rpm_expr = re.compile('rh-amazon-rhui-client-mrg-\d.*')
            elif product == 'SAP' and version == '6.5':
                rpm_expr = re.compile('rh-amazon-rhui-client-sap-hana-\d.*')
                rpm_count = 2 # exception
        elif platform == 'BETA':
            if product == 'CLOUD':
                rpm_expr = re.compile('rh-amazon-rhui-client-beta-\d.*')

        assert rpm_expr, 'spec for rhui config rpm missing: %s/%s/%s' % \
            (platform, product, version)

        # get all possible rhui config rpms installed
        rpm_log = self.get_result(connection, r'rpm -qa rh-amazon-rhui-client\*', 90)
        rpms = rpm_log.split()

        # check required rhui config rpm is installed
        assert rpms, 'no rhui config rpm installed'
        assert len(rpms) == rpm_count, 'expecting %s config rpm(s); got: %s instead' % \
            (rpm_count, rpms)
        # match expected pattern against detected rpms
        assert any([rpm_expr.match(rpm) for rpm in rpms]), \
            'wrong rhui config rpm instsalled: %s; expecting: %s' % (rpms, rpm_expr.pattern)


        return self.log



""" This module contains testcase_98_package_install_pre test """
import os
from testcase import Testcase


class testcase_98_package_install_pre(Testcase):
    """
    Do package upgrate with specified package set
    """

    tags = ['preinstall_package']
    stages = ['stage0']

    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, 'yum check-update ||:', 900)
        if 'pkg' in params:
            rpmfiles = ''
            if type(params['pkg']) == str:
                pkgs_files = [params['pkg']]
            else:
                pkgs_files = params['pkg']

            checklist = []
            for pkg in pkgs_files:
                pkgbase = os.path.basename(pkg)
                connection.sftp.put(pkg, '/tmp/%s' % pkgbase)
                rpmfiles += '/tmp/%s ' % pkgbase
                # pylint: disable=C0301
                checklist.append('[ "' + pkgbase + '" = "$(rpm -q --queryformat %{NAME}-%{VERSION}-%{RELEASE}.%{ARCH}.rpm $(rpm -qp --queryformat %{NAME} /tmp/' + pkgbase + '))" ]')
                self.get_return_value(connection, 'ls -l /tmp/%s' % pkgbase)
            self.get_return_value(connection, 'yum -y --nogpgcheck localinstall %s' % rpmfiles, 900)

            for check in checklist:
                self.get_return_value(connection, check)

        else:
            self.log.append({'result': 'skip',
                             'comment': 'no package in parameters, no upgrade was done'})
        return self.log



""" This module contains testcase_190_cds_hostnames"""
from testcase import Testcase

class testcase_190_cds_hostnames(Testcase):
    """
    ensure cdses are resolved properly
    """
    stages = ['stage1']
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """
        # contains items such as rhui2-cds02.eu-1-central.aws.ce.redhat.com
        cds_hostnames = [ "rhui2-cds0%s.%s.aws.ce.redhat.com" % (i, params['region']) \
            for i in range(1, 3)]
        # try resolving
        for hostname in cds_hostnames:
            self.get_result(connection, 'getent hosts %s' % hostname, timeout=2)
        return self.log



""" This module contains testcase_20_auditd test """
from testcase import Testcase
from distutils.version import LooseVersion


class testcase_20_auditd(Testcase):
    """
    Check auditd:
    - service should be on
    - config files shoud have specified checksums
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': r'5\..*|6\..*|7\..*'}
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """
        version = LooseVersion(params['version'])

        if version < '6.0':
            # RHEL5.x
            auditd_rules_checksum = 'f9869e1191838c461f5b9051c78a638d'
            auditd_checksum = '612ddf28c3916530d47ef56a1b1ed1ed'
            auditd_sysconf_checksum = '123beb3a97a32d96eba4f11509e39da2'
            self.ping_pong(connection, 'md5sum /etc/sysconfig/auditd  | cut -f 1 -d \' \'', auditd_sysconf_checksum)
        elif version < '7.0':
            # RHEL6.x
            auditd_rules_checksum = 'f9869e1191838c461f5b9051c78a638d'
            if '6.0' <= version <= '6.5':
                if version <= '6.1':
                    # RHEL6.0, .1
                    auditd_checksum = '612ddf28c3916530d47ef56a1b1ed1ed'
                    auditd_sysconf_checksum = '123beb3a97a32d96eba4f11509e39da2'
                else:
                    # RHEL6.2, .3, .4, .5
                    auditd_checksum = 'e1886162554c18906df2ecd258aa4794'
                    auditd_sysconf_checksum = 'd4d43637708e30418c30003e212f76fc'
            else:
                # RHEL6.6, ...
                auditd_checksum = 'e1886162554c18906df2ecd258aa4794'
                auditd_sysconf_checksum = '0825f77b49a82c5d75bcd347f30407ab'
            self.ping_pong(connection, 'md5sum /etc/sysconfig/auditd  | cut -f 1 -d \' \'', auditd_sysconf_checksum)
        elif version < '8.0':
            # RHEL7.x
            auditd_checksum = 'e1886162554c18906df2ecd258aa4794'
            auditd_rules_checksum = 'd5985c09d6c150e433362eca9d59e8fe'
        self.ping_pong(connection, 'md5sum /etc/audit/auditd.conf | cut -f 1 -d \' \'', auditd_checksum)
        self.ping_pong(connection, 'md5sum /etc/audit/audit.rules | cut -f 1 -d \' \'', auditd_rules_checksum)
        return self.log



#!/usr/bin/env python
from collections import OrderedDict

def key_parts(key):
    '''
    return list of all prefix/suffix pairs; if key == 'a.b.c':
    [
       ('a.b.c', ''),
       ('a.b', 'c'),
       ('a', 'b.c')
    ]
    '''
    sub_items = key.split('.')
    return [
        ('.'.join(list_prefix), '.'.join(sub_items[i:])) \
        for list_prefix in [sub_items[:i] for i in range(len(sub_items), 0, -1)]
    ]

class Namespace(dict):
    '''an attribure-access dictionary'''
    def __init__(self, *args, **kvargs):
        super(dict, self).__init__(*args, **kvargs)
        # a little magic
        self.__dict__ = self

    def copy(self):
        '''return a namespace made out of self'''
        return load_ns(super(Namespace, self).copy())

    def __getitem__(self, key):
        def locate(ns, key):
            if not type(key) in (unicode, str):
                # no point in splitting; try directly
                return  dict.__getitem__(ns, key)

            if key == '':
                # item found
                return ns

            if type(ns) not in (Namespace, dict):
                # item not found
                raise KeyError(repr(key))

            for item_prefix, item_suffix in key_parts(key):
                if item_prefix in ns.keys():
                    return locate(dict.__getitem__(ns, item_prefix), item_suffix, )

            # item not found
            raise KeyError(repr(key))
        return locate(self.__dict__, key)

    def __contains__(self, key):
        try:
            self[key]
            return True
        except KeyError:
            return False




def load_ns(d, leaf_processor=lambda x: x):
    '''a recursive dict-to-Namespace loader'''
    if not isinstance(d, dict):
        if isinstance(d, list):
            return [load_ns(item) for item in d]
        if isinstance(d, tuple):
            return tuple([load_ns(item) for item in d])
        return leaf_processor(d)
    ns = Namespace()
    for k, v in d.items():
        ns[k] = load_ns(v, leaf_processor)
    return ns


def dump_ns(ns, leaf_processor=lambda x: x):
    '''a recursive namespace-to-dict dumper'''
    if isinstance(ns, tuple):
        return tuple([dump_ns(item) for item in ns])
    if isinstance(ns, list):
        return [dump_ns(item) for item in ns]
    if isinstance(ns, dict) or isinstance(ns, Namespace):
        return {k: dump_ns(v) for k, v in ns.items()}

    return leaf_processor(ns)

def dump_ns_sorted(ns, leaf_processor=lambda x: x):
    '''a recursive namespace-to-dict dumper with in-place sorting'''
    if isinstance(ns, tuple):
        return tuple([dump_ns_sorted(item) for item in sorted(ns)])
    if isinstance(ns, list):
        return [dump_ns_sorted(item) for item in sorted(ns)]
    if isinstance(ns, dict) or isinstance(ns, Namespace):
        return OrderedDict([(k, dump_ns_sorted(v)) for k, v in sorted(ns.items(), key=lambda pair: pair[0])])

    return leaf_processor(ns)


def locate_ns_item(ns, item, building=False):
    '''
    locate queries of the form 'a.b.c' in a Namespace instance
    tries to find longest match
    '''
    if type(item) not in (unicode, str):
        # no point in splitting; try directly
        return  ns[item]

    if item == '':
        # item found
        return ns

    if type(ns) not in (Namespace, dict):
        # item not found
        raise KeyError(repr(item))

    for item_prefix, item_suffix in key_parts(item):
        if item_prefix in ns.keys():
            if building:
                return {item_prefix: locate_ns_item(ns[item_prefix], item_suffix, building=building)}
            else:
                return locate_ns_item(ns[item_prefix], item_suffix, building=building)

    # item not found
    raise KeyError(repr(item))


def in_ns(ns, item):
    try:
        locate_ns_item(ns, item)
        return True
    except KeyError:
        return False

def setattr_ns(obj, ns, leaf_processor=lambda x: x):
    '''kinda recursive setattr from a namespace to an arbitrary object'''

    if not isinstance(ns, dict):
        obj = leaf_processor(ns)

    for k, v in ns.items():
        try:
            setattr_ns(getattr(obj, k), v, leaf_processor=leaf_processor)
        except AttributeError:
            setattr(obj, k, leaf_processor(v))



'''
the test execution module
'''


import sys
import time
import logging
import traceback
from stitches import Expect
from ..tools.registry import TEST_CLASSES, TEST_STAGES
from ..tools.logged import logged
from ..connection.cache import get_connection, assert_connection, connection_cache_key, drop_connection, ConnectionCacheError
from data import brief
from params import when_enabled
from params import reload  as params_reload
from common import RESULT_ERROR, RESULT_PASSED, RESULT_FAILED, RESULT_SKIP, RESULT_WAIVED
from ..tools.retrying import retrying
from ..connection.contextmanager import connection as connection_ctx
from ..connection.contextmanager import alive_connection
from ..test.testcase import SkipException

TEST_WORKER_POOL_SIZE = 10 # the default SSH MaxSessions value

logger = logging.getLogger(__name__)

class TestingError(Exception):
    '''Some testing issue appeared'''

@when_enabled
def test_execute(params):
    """
    Testing stage: perform all tests required in params
    @param params: testing parameters
    @type params: dict
    @return: list of test results
    """
    assert 'test' in params, 'test field missing in %s' % brief(params)
    assert 'name' in params['test'], 'test name field missing in %s' % brief(params)
    assert 'stage' in params['test'], 'test stage field missing in %s' % brief(params)
    test_name = params['test']['name']
    test_stage = params['test']['stage']
    params['test']['exception'] = None
    params['test']['log'] = []
    params['test']['result'] = RESULT_PASSED
    hostname = params['hostname']

    logger.debug('trying %s %s %s', hostname, test_stage, test_name)
    try:
        test_cls = TEST_CLASSES[test_name]
        stage = TEST_STAGES[test_stage]
        assert test_name in stage, 'could not locate test %s in stage %s' % (test_name, test_stage)
    except KeyError as err:
        params['test_result'] = 'error: missing test/stage: %s/%s' % (test_name, test_stage)
        raise TestingError('missing test: %' % test_name)

    # asserts connection; tries reconnecting
    # FIXME: data race condition somewhere resets the ssh user
    # can't locate atm --- enforcing root /me ashamed
    params = params.copy()
    params['ssh']['user'] = 'root'
    con = get_connection(params)
    params['test']['start_time'] = time.time()
    # perform the testing
    try:
        test_obj = test_cls()
        test_obj.test(con, params)
        logger.debug('%s %s %s succeeded', hostname, test_stage, test_name)
    except AssertionError as err:
        # not caught in the test case but means the test failed
        params['test']['result'] = RESULT_FAILED
        params['test']['exception'] = traceback.format_exc()
    except SkipException as err:
        # risen by a test case, means this test is skipped
        params['test']['result'] = RESULT_SKIP
        params['test']['exception'] = traceback.format_exc()
    else:
        # no assertion errors detected --- check all cmd logs
        test_cmd_results = [cmd['result'] for cmd in test_obj.log if 'result' in cmd]
        test_result = RESULT_FAILED in test_cmd_results and RESULT_FAILED or RESULT_PASSED
        test_result = RESULT_ERROR in test_cmd_results and RESULT_ERROR or test_result
        params['test']['result'] = test_result
    finally:
        params['test']['log'] = test_obj.log
        params['test']['end_time'] = time.time()
    return params

@when_enabled
def reboot_instance(params):
    '''call a reboot'''
    with alive_connection(params) as connection:
        Expect.expect_retval(connection, 'nohup sleep 1s && nohup reboot &')
    time.sleep(10)

@when_enabled
@retrying(maxtries=60, sleep=3, loglevel=logging.DEBUG)
def wait_boot_instance(params):
    '''wait till the instance boots'''
    params_reload(params)
    _, host, user, ssh_key = connection_cache_key(params)
    with connection_ctx(host, user, ssh_key) as connection:
        assert_connection(connection)

def process_dependencies(stage_name, test_names):
    '''process process dependencies for particular stage name; yields etages of dep tree'''
    from ..tools.dependency import Graph, Vertex, bfs
    # build the graph
    root = Vertex(value='__root__')
    graph = Graph(vertices=set([root]))
    for test_name in test_names:
        logger.debug('processing testcase %s dependencies', test_name)
        test_vertex = Vertex(value=test_name)
        dep_names = getattr(TEST_CLASSES[test_name], 'after',  [])
        # in case no dependencies were specified, the test depends on the root vertice
        deps = [Vertex(dep_name) for dep_name in dep_names if dep_name in test_names] or [root]
        for dep in deps:
            logger.debug('adding dependency %s, %s', dep, test_vertex)
            graph.add_edge(dep, test_vertex)

    # evaluate dependencies; skip the root-only etage
    ret = []
    for vertex_etage in bfs(graph, root):
        test_names = [vertex.value for vertex in vertex_etage if vertex.value != '__root__']
        # FIXME: not sure why yield test_names does mess with the greenlets
        ret.append(test_names)
    return ret



def execute_tests(original_params, stage_name, pool_size=TEST_WORKER_POOL_SIZE, sorted_mode=False):
    '''perform all tests; if sorted_mode, test are sorted by both dependencies and names'''
    from gevent.pool import Pool
    pool = Pool(size=pool_size)
    for test_etage in process_dependencies(stage_name, original_params['test_stages'][stage_name]):
        for result in pool.map(test_execute, [dict(test=dict(name=test_name, stage=stage_name), **original_params) \
                for test_name in sorted_mode and sorted(test_etage) or test_etage]):
            yield result

def execute_stages(params, pool_size=TEST_WORKER_POOL_SIZE, sorted_mode=False):
    '''perform all stages; sorted_mode, pool_size are propagated to `execute_tests`'''
    params['stage_name'] = 'execute_tests'
    unixtime = time.time()
    params['start_time'] = unixtime
    params['end_time'] = unixtime # end time in this case does not make sense as results are same copies in paralel run
    stages = sorted(params['test_stages'])
    # reboots inbetween stages
    for stage_name in stages[:-1]:
        if not params['test_stages'][stage_name]:
            logger.debug('skipping empty test stage: %s', stage_name)
            continue # avoid rebooting on empty stages
        for result in execute_tests(params, stage_name, pool_size=pool_size, sorted_mode=sorted_mode):
            yield result
        reboot_instance(params)
        wait_boot_instance(params)
    # no reboot at last stage exit
    for result in execute_tests(params, stages[-1]):
        yield result



""" This module contains testcase_03_running_services test """

from testcase import Testcase

import yaml


class testcase_03_running_services(Testcase):
    """
    Check for running services
    """
    stages = ['stage1']
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        with open(self.datadir + '/running_services.yaml') as expected_services_fd:
            all_services = yaml.safe_load(expected_services_fd)
        try:
            expected_services = all_services['%s_%s' % (prod, ver)]
        except KeyError:
            self.log.append({
                'result': 'skip',
                'comment': 'unsupported region and/or platform-version combination'})
            return self.log

        is_systemd = self.get_result(connection, 'rpm -q systemd > /dev/null && echo True || echo False')
        if is_systemd == 'True':
            for service in expected_services:
                self.get_return_value(connection, 'systemctl is-active %s.service' % service)
        else:
            for service in expected_services:
                self.ping_pong(connection, 'chkconfig --list %s' % service, '3:on')
        return self.log



'''
tool for uploading data to Jenkins
'''
import sys
import time
import logging
import tempfile
import json
import requests
import time
import urllib3
import StringIO
import xml.etree.ElementTree as ET
from ..tools.retrying import retrying, EAgain
from ..work.data import load_yaml, save_result
from ..work.common import RESULT_PASSED

logger = logging.getLogger(__name__)

def main(config, istream, job, hashtag, desc, login, passwd, report_empty=False):
    file_data = istream.read()
    # check the data
    report_tree = ET.parse(StringIO.StringIO(file_data))
    if not report_tree.getroot().getchildren() and not report_empty:
        # nothing to report and user not interested in reporting empty result
        print "nothing to report..."
        return
    params = {"parameter": [{'name':'report.xml', 'file': 'file0'},{"name": "desc", "value": hashtag}]}
    data, content_type = urllib3.encode_multipart_formdata([
        ("file0", (istream.name, file_data)),
        ("json", json.dumps(params)),
        ("Submit", "Build"),
        ])
    url = job + 'build'
    r = requests.post(url, auth=(login, passwd), data=data, headers={"content-type": content_type})
    if not r.ok:
        print "Authentification failure."
        exit(1)
    print "Waiting 10s for build to be finished."
    time.sleep(10)
    url = job + 'api/xml'
    xpath_str = "//build//*[contains(text(),'"+ hashtag +"')]/../../../url"
    params = {'depth': '1', 'xpath': xpath_str, 'wrapper': 'list'}
    r = requests.get(url, auth=(login, passwd), params=params)
    
    tree = ET.fromstring(r.text.replace('</url>','</url>\n'))
    notags = ET.tostring(tree, encoding='utf8', method='text')
    job_number = notags.splitlines()[0].rsplit('/',2)[1]

    url = job + 'default/' + job_number + '/testReport/submitDescription'
    params = {'description': desc}
    r = requests.get(url, auth=(login,passwd), params=params)

    print "Return code: " + str(r.status_code)



""" This module contains testcase_24_yum_plugin test """
from testcase import Testcase


class testcase_24_yum_plugin(Testcase):
    """
    RHN plugin should be disabled
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (>=5.5, <7.0)'}
    after = ['testcase_11_package_set']
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, 'grep \'^enabled[[:space:]]*=[[:space:]]*[^0 ]\' /etc/yum/pluginconf.d/rhnplugin.conf', expected_status=1)
        return self.log



""" This module contains testcase_08_memory test """
from testcase import Testcase
import re


class testcase_08_memory(Testcase):
    """
    Check for available amount of memory
    """
    stages = ['stage1', 'stage2']
    tags = ['default', 'kernel']

    def test(self, connection, params):
        """ Perform test """

        if not 'memory' in params.keys():
            self.log.append({'result': 'skip', 'comment': 'memory parameter is not set, nothing to check'})
        else:
            values = self.match(connection, 'grep --color=none \'MemTotal:\' /proc/meminfo', re.compile(r'.*\r\nMemTotal:\s*([0-9]+) ', re.DOTALL))
            if values:
                self.get_return_value(connection, '[ %s -gt %s ]' % (values[0], params['memory']))
        return self.log



'''
basic serial test processor
'''
import time
import logging
import traceback
from stage import assert_image, create_instance, attempt_ssh, allow_root_login, global_setup_script, terminate_instance
from test import execute_stages, TEST_WORKER_POOL_SIZE
from common import RESULT_ERROR
from stage import STAGES, StageError, SkipError


logger = logging.getLogger(__name__)


def process(params, pool_size=TEST_WORKER_POOL_SIZE, sorted_mode=False):
    '''process required acctions'''
    terminate = False
    try:
        params = assert_image(params)
        yield params
        params = create_instance(params)
        terminate = True
        yield params
        params = attempt_ssh(params)
        yield params
        params = allow_root_login(params)
        yield params
        params = global_setup_script(params)
        yield params
        for test_result in execute_stages(params, pool_size=pool_size, sorted_mode=sorted_mode):
            yield test_result
    except SkipError as err:
        logger.debug('encountered skip error: %s', err)
        yield params
    except StageError as err:
        logger.debug('encountered stage error: %s', err)
        yield params
    except Exception as err:
        # unhandled error
        logger.error('unhandled exception: %s(%s)', type(err), err)
        params['stage_exception'] = traceback.format_exc()
        params['stage_result'] = RESULT_ERROR
        yield params
    finally:
        if terminate:
            yield terminate_instance(params)

def required_actions_count(params):
    total = 0
    for item in params:
        total += len(STAGES)
        for test_stage in item['test_stages']:
            for test_name in item['test_stages'][test_stage]:
                total += 1
    return total

def print_progress_info(actual, total):
    '''print progress info actual/total'''
    print '# %s %2.2f%% (%s/%s)' % (time.ctime(), total and float((actual * 100)/total) or 0, actual, total)





""" This module contains testcase_34_cpu test """
from testcase import Testcase


class testcase_34_cpu(Testcase):
    """
    Check the number of cpu cores available
    """
    stages = ['stage1']
    tags = ['default', 'kernel']

    def test(self, connection, params):
        """ Perform test """

        if 'cpu' in params.keys():
            self.ping_pong(connection, 'cat /proc/cpuinfo | grep \'^processor\' | wc -l', params['cpu'])
        else:
            self.log.append({'result': 'skip',
                             'comment': 'cpu in hwp is not set'
                             })
        return self.log



""" Result parsing functions """
import textwrap
import aggregate
from ..work.common import RESULT_PASSED, RESULT_FAILED, RESULT_ERROR, RESULT_SKIP, RESULT_WAIVED

COMMON_COMMAND_KEYS_ORDERED=('command', 'match', 'result', 'value', 'actual', 'expected', 'comment')

def command_repr(command):
    '''repr a single command as list of lines'''
    ret = []
    format_value = lambda key, value: textwrap.wrap(('%s: ' % key) + str(value), initial_indent='  ',
                        subsequent_indent='  ', break_on_hyphens=False,  break_long_words=True, width=70)
    ret.append('-')
    for key in COMMON_COMMAND_KEYS_ORDERED:
        if key not in command:
            continue
        ret.extend(format_value(key, command[key]))
    for key in set(command) - set(COMMON_COMMAND_KEYS_ORDERED):
        ret.extend(format_value(key, command[key]))
    return ret


def get_test_result(test_data, whitelist=[], verbose=False):
    '''get formated test result'''
    try:
        ret = test_data['result']
    except KeyError:
        # no result -- test skipped
        ret = RESULT_SKIP
    try:
        test_log = test_data['log']
    except KeyError:
        # no test log
        test_log = []
    if test_data['name'] in whitelist:
        ret = RESULT_WAIVED
    log = ['%s:%s: %s' % (test_data['stage'], test_data['name'], ret)]
    if 'exception' in test_data and test_data['exception'] and (verbose or \
            ret in [RESULT_FAILED, RESULT_ERROR]):
        log.append(test_data['exception'])
    if ret != RESULT_PASSED or verbose:
         for command in test_log:
            log.extend(command_repr(command))
    return ret, log


def get_stage_result(stage_data, verbose=False):
    '''get formated stage result'''
    ret = stage_data['stage_result']
    log = []
    if ret != RESULT_PASSED or verbose:
        log = ['-']
        txt = '%s: %s\n%s' % (stage_data['stage_name'], stage_data['stage_result'], stage_data['stage_exception'])
        log.extend(textwrap.wrap(txt, break_on_hyphens=False, break_long_words=True, width=70, initial_indent='  ',
                                    subsequent_indent='  '))
    return ret, log


def get_hwp_result(data, whitelist=[], verbose=False):
    '''get overal hwp result'''
    ret = RESULT_PASSED
    log = []
    for res in data:
        if 'test' in res:
            # test case result
            sub_result, sub_log = get_test_result(res['test'], whitelist, verbose)
        else:
            # stage result
            sub_result, sub_log = get_stage_result(res, verbose)
        if sub_result in [RESULT_ERROR, RESULT_FAILED, RESULT_SKIP,RESULT_WAIVED] and ret == RESULT_PASSED:
            ret = sub_result
        log.extend(sub_log)
    return ret, log

def get_ami_result(data, whitelist, verbose=False):
    '''get overal ami result'''
    ret = RESULT_PASSED
    log = []
    for hwp in data:
        hwp_index = len(log)
        sub_result, sub_log = get_hwp_result(data[hwp], whitelist, verbose)
        if sub_result not in [RESULT_PASSED, RESULT_SKIP] and ret == RESULT_PASSED:
           ret = sub_result
        header = '%s: %s' % (hwp, sub_result)
        log.insert(hwp_index, '-' * len(header))
        log.insert(hwp_index, header)
        log.insert(hwp_index, '')
        log.extend(sub_log)
    return ret, log

def get_overall_result(data, whitelist=[], verbose=False):
    """
    Get human-readable representation of the result; partitioned by ami
    returns a tuple of an overal result and list of tuples overal_result, [(ami_resutl, ami_log), ...]
    """
    ret = RESULT_PASSED
    log = []
    if not data:
        return ret, log

    agg_data = aggregate.nested(data, 'region', 'arch', 'itype', 'ami', 'cloudhwname')
    for region in agg_data:
        for arch in agg_data[region]:
            for itype in agg_data[region][arch]:
                for ami in agg_data[region][arch][itype]:
                    sub_result, sub_log = get_ami_result(agg_data[region][arch][itype][ami], whitelist, verbose)
                    if sub_result != RESULT_PASSED and ret == RESULT_PASSED:
                        ret = sub_result
                    ami_header = '%s %s %s %s: %s' % (region, arch, itype, ami, ret)
                    sub_log.insert(0, '-' * len(ami_header))
                    sub_log.insert(0, ami_header)
                    log.append((sub_result, sub_log))
    return ret, log



from ec2 import EC2
from openstack_ec2 import OpenStackEC2
from base import PermanentCloudException, AbstractCloud

def get_driver(cloud_name, logger, maxwait):
    driver_cls = AbstractCloud
    driver_clses = [driver_cls for driver_cls in [EC2, OpenStackEC2, AbstractCloud] if \
                        getattr(driver_cls, 'cloud', None) == cloud_name]
    if not driver_clses:
        raise PermanentCloudException('no driver for cloud name: %s' % cloud_name)
    driver_cls = driver_clses[0]
    driver = driver_cls(logger, maxwait)
    return driver

__all__ = ['get_driver']



""" This module contains testcase_98_kernel_parameter test """
import os
import paramiko
import time
from testcase import Testcase


class testcase_98_kernel_parameter(Testcase):
    """
    Add specific kernel parameters
    """

    tags = []
    stages = ['stage0']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        if 'kernelparams' in params:
            self.get_return_value(connection, 'sed -i \'s,\\(kernel .*$\\),\\1 %s,\' /boot/grub/grub.conf' % params['kernelparams'])
            self.get_return_value(connection, 'nohup sleep 1s && nohup reboot &')
            time.sleep(30)
        else:
            self.log.append({'result': 'skip',
                             'comment': 'no kernelparam was provided'})
        return self.log



""" This module contains testcase_55_yum_group_install test """
from testcase import Testcase


class testcase_55_yum_group_install(Testcase):
    """
    Try to install 'Development tools' group with yum
    """
    stages = ['stage1']
    after = ['testcase_50_yum_package_install']
    tags = ['default', 'content']
    applicable = {"platform": "(?i)RHEL|BETA", "version": "OS (>=5.5, !=6.0)"}

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, 'yum -y groupinstall \'Development tools\'', 1200)
        # Checking whether something was installed
        self.get_return_value(connection, 'rpm -q glibc-devel')
        return self.log



import os
import re
import random
import logging
from collections import defaultdict
from ..tools.registry import TEST_STAGES, TEST_CLASSES
from ..tools.namespace import load_ns, dump_ns
from tags_filter import factory as applicability_factory
from functools import wraps

from yaml import load as yload
from yaml import dump_all as ysafe_dump_all
try:
    from yaml import CSafeLoader as Loader, CSafeDumper as Dumper
except ImportError:
    from yaml import SafeLoader as Loader, SafeDumper as Dumper


logger = logging.getLogger(__name__)
HWP_PATHS = ['./hwp/', '/usr/share/dva/hwp']
CONFIG_FILES = ['dva.yaml', '~/.dva.yaml', '/etc/dva.yaml']

DEFAULT_FIELDS = {
    'cloud': 'ec2',
    'enabled': True,
    'bmap': {},
    'userdata': '',
    'keepalive': False,
    'region_endpoint': None
}

EPHEMERAL_FIELDS = ['credentials', 'global_setup_script', 'instance', 'ssh', 'test_stages', 'user_data',
        'bmap', 'enabled', 'userdata', 'test_whitelist', 'test_blacklist', 'stage_whitelist', 'stage_blacklist',
        'tags_whitelist', 'tags_blacklist']

GLOBAL_CONFIG_FIELDS = ['global_setup_script']

RE_ALL = re.compile('.*')

class DataError(Exception):
    '''data wasn't OK'''

class ConfigError(Exception):
    '''configuration wasn't OK'''

def load_yaml(filename):
    ''' load yaml file'''

    if isinstance(filename, file):
        return yload(filename, Loader=Loader)

    with open(filename) as datafd:
        data = yload(datafd, Loader=Loader)
    return data

def arch_to_path(arch):
    '''return path to first arch if found in HWP_PATHS; raise otherwise'''
    if os.path.exists(arch):
        # use direct custom hwp path from data file instead of an alias
        return arch
    filename = lambda dirname: os.path.join(dirname, str(arch) + '.yaml')
    arch_files = [filename(dirname) for dirname in HWP_PATHS if os.path.exists(filename(dirname))]
    assert arch_files, 'wrong arch: %s.yaml not found in any of %s' % (arch, HWP_PATHS)
    logger.debug('using hwp(arch) file: %s', arch_files[0])
    return arch_files[0]

def set_config_filename(config_file=None):
    if config_file is None:
        # no config file given, try locating some in default spots
        config_files = [filename for filename in CONFIG_FILES if os.path.exists(filename)]
        assert config_files, 'could not locate any config file; tried: %s' % CONFIG_FILES
        config_file = config_files[0]

    logger.debug('using config file: %s', config_file)
    return config_file


def record_cloud_config(record, config_file=None):
    '''put cloud config data into a record; first found config file is used'''
    config_file = set_config_filename(config_file)
    # e.g. record['cloud'] == 'ec2' -> config['cloud_access']['ec2']
    global_config = cloud_config = {}
    try:
        region = record['region']
        cloud = record['cloud']
    except KeyError as err:
        raise DataError('record %r misses %s' % (record, err))
    try:
        global_config = load_yaml(config_file)
        cloud_config = global_config['cloud_access'][cloud]
        # provide cloud credentials
        record['credentials'] = {
            'ec2_access_key': cloud_config['ec2_access_key'],
            'ec2_secret_key': cloud_config['ec2_secret_key']
        }
        # provide ssh config
        ssh_config = cloud_config['ssh'][region]
        record['ssh'] = {
            'keypair': ssh_config[0],
            'keyfile': ssh_config[1]
        }
    except KeyError as err:
        raise ConfigError('config %s(:ssh) misses %s' % (config_file, err))
    except IndexError as err:
        raise ConfigError('config %s:ssh.%s misses an item (%s)' % (config_file, region, err))
    except Exception as err:
        raise ConfigError('config %s: %s' % (config_file, err))

    # propagate global config fileds
    for key in GLOBAL_CONFIG_FIELDS:
        try:
            record[key] = global_config[key]
        except KeyError:
            pass
    return record


def expand_record_arch(record):
    '''return a list of arch-expanded records'''
    assert type(record) is dict
    try:
        arch = record['arch']
    except KeyError as err:
        raise DataError('record %r: %s' % (record, err))
    if isinstance(arch, basestring):
        # possibly a path or an HWP alias
        arch_data = load_yaml(arch_to_path(arch))
    elif isinstance(arch, list):
        # possibly embedded data
        arch_data = arch
    else:
        raise DataError('record %r: wrong arch type: %s (should be list of dicts)' % (record, type(arch)))
    return [dict(record.items() + arch_item.items()) for arch_item in arch_data]

def expand_record_tests(record):
    '''
    expand record with test-case/stage lists:
    - adds record['test_stages'] what is a dictionary of lists of test_names
    - e.g. {'stage1': [test_a, test_b, ...], 'stage2': [test_x, test_y, ...], ...}
    takes care of applicability and black/white-listing
    '''
    assert type(record) is dict
    app_filter = lambda test_class: getattr(test_class, 'applicable', defaultdict(lambda: None))
    not_filter = lambda test_class: getattr(test_class, 'not_applicable', defaultdict(lambda: None))
    tags = lambda test_class: getattr(test_class, 'tags', ['default'])
    record['test_stages'] = \
    { stage: \
        [
            test_name for test_name in sorted(TEST_STAGES[stage]) if \
                applicability_factory(applicable=app_filter(TEST_CLASSES[test_name]), \
                                        not_applicable=not_filter(TEST_CLASSES[test_name]))(record) and \
                any([re.match(pattern, test_name) for pattern in record.get('test_whitelist', ['.*'])]) and \
                not any([re.match(pattern, test_name) for pattern in record.get('test_blacklist', [])]) and \
                any([re.match(pattern, tag) for tag in tags(TEST_CLASSES[test_name]) \
                    for pattern in record.get('tags_whitelist', ['default'])]) and \
                not any([re.match(pattern, tag) for tag in tags(TEST_CLASSES[test_name]) \
                    for pattern in record.get('tags_blacklist', [])])
        ]
        for stage in sorted(TEST_STAGES) if \
            any([re.match(pattern, stage) for pattern in record.get('stage_whitelist', ['.*'])]) and \
            not any([re.match(pattern, stage) for pattern in record.get('stage_blacklist', [])])
    }
    return record

def normalize_record(record):
    '''return a record with normalized fields'''
    assert type(record) is dict, 'invalid record type: %s' % type(record)
    # make sure all is lowercase and "safe"
    tmp = load_ns(record, leaf_processor=lambda value: type(value) is str and value.lower() or value)
    record = dump_ns(tmp)
    # override any defaults with original record
    # to prevent any missing fields
    return dict(DEFAULT_FIELDS.items() + record.items())


def expand_record(record, config_file, augment={}):
    '''
    expand single validation record record according to configuration data
    return: list of action items for a single record
    '''
    record.update(augment)
    return [expand_record_tests(record) for record in expand_record_arch( \
                    record_cloud_config(normalize_record(record), config_file))]

def load(path, config_file=None, augment={}, shuffle=True):
    '''load data and process it; returns list of expanded records
       all **kvs are propagated to the data
    '''

    data = load_yaml(path)
    if data is None:
        # no items to process
        return []

    assert type(data) is list, 'invalid data type: %s' % type(data)
    ret = []
    for record in data:
        ret.extend(expand_record(record, config_file, augment))
    # shuffle to avoid single region exhaustion
    if shuffle:
        random.shuffle(ret)
    return ret

def save_result(stream, result):
    '''shortcut for ydump_all(data, stream, Dumper=Dumper)'''
    #result = dump_ns(result, leaf_processor=lambda value: type(value) is unicode and str(value) or value)
    if isinstance(stream, file):
        ysafe_dump_all([[result]], stream, Dumper=Dumper)
    else:
        with open(stream, 'w') as fd:
            ysafe_dump_all([[result]], fd, Dumper=Dumper)



def brief(record):
    '''return a brief info about the record'''
    return '%(region)s %(ami)s %(cloudhwname)s %(platform)s %(version)s %(itype)s' % record


def strip_ephemeral(record):
    '''strip ephemeral record fields such as host names, instance details, authentication'''
    ret = record.copy()
    for field_name in EPHEMERAL_FIELDS:
        try:
            del(ret[field_name])
        except KeyError:
            pass
    return ret



""" This module contains testcase_98_kernel_upgrade_pre test """
import os
import paramiko
import time
from testcase import Testcase


class testcase_98_kernel_upgrade_pre(Testcase):
    """
    Do kernel upgrate (with specified package set or from repo) before testing
    """

    tags = ['kernel', 'preinstall_kernel']
    stages = ['stage0']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        if (prod in ['RHEL', 'BETA']) and (ver.startswith('6.')):
            # remove bfa-firmware - conflicts with new RHEL6 kernel
            self.get_return_value(connection, 'rpm -q bfa-firmware && rpm -e bfa-firmware ||:', 60)
        if 'kernelpkg' in params:
            kernelfiles = ''
            if type(params['kernelpkg']) == str:
                pkgs_files = [params['kernelpkg']]
            else:
                pkgs_files = params['kernelpkg']
            for pkg in pkgs_files:
                pkgbase = os.path.basename(pkg)
                connection.sftp.put(pkg, '/tmp/%s' % pkgbase)
                kernelfiles += '/tmp/%s ' % pkgbase
                self.get_return_value(connection, 'ls -l /tmp/%s' % pkgbase)
            if len(pkgs_files) == 1:
                self.get_return_value(connection, 'yum -y update %s' % kernelfiles, 900)
                kernel_updated = self.get_return_value(connection, '[ $(rpm -qa kernel | wc -l) -gt 1 ]', nolog=True)
            else:
                kernel_updated = self.get_return_value(connection, 'rpm -U %s' % kernelfiles, 900)
        else:
            # doing upgrade from repo
            self.get_return_value(connection, 'yum -y install kernel', 300)
            kernel_updated = self.get_return_value(connection, '[ $(rpm -qa kernel | wc -l) -gt 1 ]', nolog=True)
        if kernel_updated == 0:
            if prod == 'FEDORA' and ver == '18':
                # we have a bug in kernel upgrade
                # pylint: disable=C0301
                self.get_result(connection, 'cat /boot/grub/grub.conf | sed -e \'s|hd0,0|hd0|\' -e \'s|default=1|default=0|\' > /boot/grub/menu.lst && echo SUCCESS')
            self.get_return_value(connection, 'nohup sleep 1s && nohup reboot &')
            time.sleep(30)
        else:
            self.log.append({'result': 'skip',
                             'comment': 'no kernel upgrade was done'})
        return self.log



""" This module contains testcase_29_swap_file test """
from testcase import Testcase


class testcase_29_swap_file(Testcase):
    """
    Swap should be enabled and it should be possible to do swapoff/swapon
    (not applicable for t1.micro and hvm instances)
    """
    stages = ['stage1']
    applicable = {'cloudhwname': r'(?!t1\.micro)', 'virtualization': '(?!hvm)', 'arch': '(?!x86_64)'}
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        size = self.get_result(connection, 'parted -l | grep linux-swap | awk \'{print $4}\' | awk -F\'MB\' \'{print $1}\'', 15)
        partition = self.get_result(connection, 'parted -l | grep -B 6 swap | grep \'^Disk /\' | awk \'{print $2}\' | sed \'$s/.$//\' | head -1', 15)
        if size and partition:
            self.get_return_value(connection, '[ ' + size + ' -gt 0 ]')
            self.get_return_value(connection, 'swapoff ' + partition + ' ; echo')
            self.get_return_value(connection, 'swapon ' + partition, 30)
            self.get_return_value(connection, 'swapoff ' + partition + ' && swapon ' + partition, 30)
        return self.log



""" This module contains testcase_01_bash_history test """
from testcase import Testcase
from dva.work.stage import OLD_BASH_HISTORY_FILE as HISTFILE, SSH_USERS
from os import path


class testcase_01_bash_history(Testcase):
    """
    Ensure /root/.bash_history file is empty
    """
    stages = ['stage1']
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """
        # try all possible cloud users
        for user in SSH_USERS:
            # 'expand' user
            user_histfile = '~' + user + path.sep + path.basename(HISTFILE)
            self.ping_pong(connection, '[ ! -f %s ] && echo 0 || cat %s | wc -l' % (user_histfile, user_histfile), '\r\n0\r\n')

        return self.log



""" This module contains testcase_90_2_cgroups_kernel_bug914737 test """
import time
from testcase import Testcase


class testcase_90_3_bug1074660(Testcase):
    """
    Reproducer for kernel bug 1074660
    """
    stages = ['stage1']
    tags = []

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version'].upper()
        self.get_return_value(connection, 'yum -y install gcc-c++ wget', 300)
        self.get_return_value(connection, 'wget http://www.la-samhna.de/samhain/samhain_signed-2.8.6.tar.gz', 300)
        self.get_return_value(connection, 'tar -zxvf samhain_signed-2.8.6.tar.gz', 10)
        self.get_return_value(connection, 'tar -zxvf samhain-2.8.6.tar.gz', 10)
        self.get_return_value(connection, 'cd samhain-2.8.6 && ./configure && make && make install', 100)
        self.get_return_value(connection, 'samhain -t init', 300)
        self.get_return_value(connection, 'samhain -t check -D', 100)
        try:
            time.sleep(600)
            self.get_return_value(connection, 'id', 30)
        except:
            self.log.append({'result': 'failed', 'command': 'bug reproducer succeeded'})
        return self.log



"""
Basic abstract cloud class and exceptions
"""
import contextlib

class TemporaryCloudException(Exception):
    """
    Temporary exception, 'try again'
    """
    pass


class PermanentCloudException(Exception):
    """
    Permanent exception, no need to try any more
    """
    pass


class SkipCloudException(Exception):
    """
    'Skip' exception - missing hardware in the region,...
    """
    pass


class UnknownCloudException(Exception):
    """
    Unknown exception, 'try again' but increase ntry
    """
    pass


class AbstractCloud(object):
    """
    Abstract cloud class
    """
    cloud = 'default'
    mandatory_fields = []

    def __init__(self, logger, maxwait):
        self.logger = logger
        self.maxwait = maxwait

    def set_default_params(self, params, cloud_access_config):
        raise NotImplementedError

    def create(self, params):
        raise NotImplementedError

    def terminate(self, params):
        raise NotImplementedError

    def get_console_output(self, params):
        raise NotImplementedError

    def update(self, params):
        raise NotImplementedError

    def get_connection(self, params):
        raise NotImplementedError

    @contextlib.contextmanager
    def connection(self, params):
        raise NotImplementedError

    def get_image(self, params):
        raise NotImplementedError

    def get_instance(self, params):
        raise NotImplementedError



""" This module contains testcase_04_cloud_firstboot test """

from testcase import Testcase


class testcase_04_cloud_firstboot(Testcase):
    """
    Check that rh-cloud-firstboot is disabled
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (>=5.5, <7.0)'}
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.ping_pong(connection, 'chkconfig --list rh-cloud-firstboot', '3:off')
        self.get_return_value(connection, 'test -f /etc/sysconfig/rh-cloud-firstboot')
        self.ping_pong(connection, 'cat /etc/sysconfig/rh-cloud-firstboot', 'RUN_FIRSTBOOT=NO')
        return self.log



""" This module contains testcase_80_no_avc_denials test """
from testcase import Testcase


class testcase_80_no_avc_denials(Testcase):
    """
    Check for avc denials absence
    """
    tags = ['default']
    stages = ['stage1', 'stage2']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.ping_pong(connection, 'echo START; grep \'avc:[[:space:]]*denied\' /var/log/messages /var/log/audit/audit.log | grep -v userdata; echo END',
                       '\r\nSTART\r\nEND\r\n', 60)

        return self.log



""" This module contains testcase_17_shells test """
from testcase import Testcase


class testcase_17_shells(Testcase):
    """
    Check for bash/nologin shells in /etc/shells
    """
    stages = ['stage1']
    tags = ['default']

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        self.get_return_value(connection, 'grep \'bin/bash$\' /etc/shells')
        self.get_return_value(connection, 'grep \'bin/nologin$\' /etc/shells')
        return self.log



'''
the retrying callable decorator
'''
import logging
import sys
from functools import wraps
logger = logging.getLogger(__name__)

class EAgain(RuntimeError):
    '''raise to retry'''

class RetryingError(RuntimeError):
    '''too many retries'''

def retrying(maxtries=3, sleep=10, loglevel=logging.WARNING, final_exception=RetryingError):
    '''
    retrying-logic decorator
    @param maxtries: how many times before giving up
    @param sleep: how long to wait between retries
    @param loglevel: at which level to report retries
    @param final_exception: what exception breaks the retries immediately
                            same is risen when too many attempts were hit
    '''
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kvs):
            import time
            import traceback
            ntries = 0
            err = None
            while ntries < maxtries:
                ntries += 1
                logger.debug('%s: try: #%s', fn.__name__, ntries)
                try:
                    return fn(*args, **kvs)
                except EAgain as err:
                    logger.log(loglevel, '%s: (try #%s): %s --- retrying in %ss', fn.__name__, ntries, err, sleep)
                    time.sleep(sleep)
            # not reached in case fn call was successful --- return statement
            else:
                raise final_exception('%s: %s. Too many retries: %s.' % (fn.__name__, err, ntries))
        return wrapper
    return decorator


__all__ = ['retrying']



""" This module contains testcase_09_nameserver test """
from testcase import Testcase


class testcase_09_nameserver(Testcase):
    """
    Check if DNS resolving works
    """
    stages = ['stage1']
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        if prod in ['RHEL', 'BETA'] and (ver.startswith('5.') or ver.startswith('6.')):
            self.get_return_value(connection, 'dig clock.redhat.com | grep 66.187.233.4', 30)
        else:
            self.get_return_value(connection, 'ping -c 5 clock.redhat.com', 30)
        return self.log



""" This module contains testcase_90_2_cgroups_kernel_bug914737 test """
import time
from testcase import Testcase


class testcase_90_2_cgroups_kernel_bug914737(Testcase):
    """
    Reproducer for kernel bug 914737
    """
    stages = ['stage1']
    tags = ['kernel']
    not_applicable = {'platform': '(?i)ATOMIC'}

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version'].upper()
        # pylint: disable=W0702
        try:
            memory = int(params['memory'])
        except:
            self.log.append({'result': 'skip', 'command': 'memory is not set in the profile'})
            return self.log
        self.get_return_value(connection, 'if [ ! -f /bin/cgset ]; then yum -y install libcgroup-tools ; fi', 240)
        self.get_return_value(connection, 'if ! mount | grep cgroup ; then service cgconfig start ; fi')
        connection.sftp.put('/usr/share/dva/data/memhog_with_forks.c', '/root/memhog_with_forks.c')
        if prod == 'FEDORA' and ver == '18':
            # ugly workaround - dependency problems :-(
            self.get_result(connection, 'if ! rpm -q gcc ; then yum -y install glibc audit gcc; fi', 300)
        else:
            self.get_result(connection, 'if ! rpm -q gcc ; then yum -y install gcc; fi', 300)
        self.get_return_value(connection, 'gcc /root/memhog_with_forks.c -o /root/memhog_with_forks')
        # Creating cpu and memory cgroups
        self.get_return_value(connection, 'for i in `seq 0 1000 `; do cgcreate -g cpu:/Group$i ; cgcreate -g memory:/Group$i ; done')
        # pylint: disable=C0301,W0702
        self.get_return_value(connection, 'for i in `seq 0 1000 `; do cgset -r memory.limit_in_bytes=' + str(memory) + ' /Group$i ; cgset -r cpu.shares=$i /Group$i ; done')
        try:
            # pylint: disable=C0301
            self.get_result(connection, 'for i in `seq 0 1000 `; do cgexec -g cpu:/Group$i -g memory:/Group$i /root/memhog_with_forks %i & echo $i ; done' % (memory // 1000000), 60)
            time.sleep(30)
            self.get_return_value(connection, 'id', 30)
            self.get_return_value(connection, 'killall memhog_with_forks ||:', 30)
        except:
            self.log.append({'result': 'failed', 'command': 'bug reproducer succeeded'})
        return self.log



""" This module contains testcase_27_yum_repos test """
from testcase import Testcase
from dva.work.common import RESULT_FAILED, RESULT_SKIP
import yaml


class testcase_27_yum_repos(Testcase):
    """
    Check for enabled yum repos
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (>=5.5, !=6.0)'}
    after = ['testcase_24_yum_plugin']
    tags = ['default', 'content']

    def test(self, connection, params):
        """ Perform test """

        plat = params['platform'].upper()
        prod = params['product'].upper()
        ver = params['version']
        if connection.rpyc is None:
            self.log.append({
                'result': RESULT_FAILED,
                'comment': 'test can\'t be performed without RPyC connection'})
            return self.log
        repos = {}
        rbase = connection.rpyc.modules.yum.YumBase()
        for repo in rbase.repos.repos:
            repos[repo] = rbase.repos.repos[repo].isEnabled()

        # figure out whether expected repos match repos
        with open(self.datadir + '/repos.yaml') as expected_repos_fd:
            all_repos = yaml.safe_load(expected_repos_fd)
        try:
            expected_repos_ = all_repos[params['region']]['%s_%s' % (prod, ver)]
        except KeyError:
            try:
                expected_repos_ = all_repos[params['region']]['%s_%s' % (plat, ver)]
            except KeyError:
                self.log.append({
                    'result': RESULT_FAILED,
                    'comment': 'unsupported region and/or platform-version combination'})
                return self.log
        # expand %region%
        expected_repos = {}
        for key, val in expected_repos_.items():
            expected_repos[key.replace('%region%', params['region'])] = val
        ret = {
            'expected repos': expected_repos,
            'actual repos': repos
        }
        ret['result'] = expected_repos == repos and 'passed' or 'failed'
        self.log.append(ret)
        return self.log



'''
connection ctx manager module
'''

import stitches
import contextlib
import logging
from ..tools.retrying import retrying, EAgain
from cache import CONNECTION_ATTEMPTS, CONNECTION_ATTEMPTS_RETRY_AFTER, assert_connection, connection_cache_key

@contextlib.contextmanager
def connection(host, user, key):
    '''ctx manager for connections'''
    con = stitches.Connection(host, user, key)
    try:
        yield con
    finally:
        con.disconnect()
        del(con)


@contextlib.contextmanager
@retrying(maxtries=CONNECTION_ATTEMPTS, sleep=CONNECTION_ATTEMPTS_RETRY_AFTER, loglevel=logging.DEBUG)
def alive_connection(params):
    '''a context manager yielding an asserted/alive connection'''
    _, host, user, ssh_key = connection_cache_key(params)
    with connection(host, user, ssh_key) as con:
        assert_connection(con)
        yield con



""" This module contains testcase_97_collect_packagelist test """
from testcase import Testcase
import re


class testcase_97_collect_packagelist(Testcase):
    """
    Collect package list (debugging)
    """
    stages = ['stage0']
    tags = []

    # pylint: disable=W0613
    def test(self, connection, params):
        """ Perform test """

        packages = self.match(connection,
                              'rpm -qa --queryformat \'%{NAME},\' && echo',
                              re.compile('.*\r\n(.*),\r\n.*', re.DOTALL),
                              timeout=30)
        if packages is None:
            self.log.append({'result': 'failed', 'comment': 'Failed to get package set'})
        else:
            self.log.append({'result': 'passed', 'comment': packages[0]})
        return self.log



""" This module contains testcase_31_subscription_management test """
from testcase import Testcase
from distutils.version import LooseVersion

class testcase_31_subscription_management(Testcase):
    """
    Subscription manager shoud be disabled
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (> 5.5)'}
    after = ['testcase_27_yum_repos']
    tags = ['default', 'content']

    def test(self, connection, params):
        """ Perform test """
        version = LooseVersion(params['version'])
        # following versions shouldn't contain subscription-manager at all
        if version < '5.9' or '6.0' >= version < '6.3':
            self.ping_pong(
                connection,
                'rpm -q subscrioption-manager || echo SUCCESS',
                expectation='\r\nSUCCESS\r\n',
            )
            return self.log

        # check subscription-manager plugin is disabled
        self.ping_pong(
            connection,
            'yum --disablerepo=\'*\' -v repolist',
            expectation='Not loading "subscription-manager" plugin',
            timeout=30
        )
        # check subscription-manager plugin can be enabled
        self.ping_pong(
            connection,
            'yum --enableplugin=subscription-manager --disablerepo=\'*\' -v repolist',
            expectation='Loading "subscription-manager" plugin',
            timeout=30
        )
        if params['version'].startswith('5.'):
            # RHEL5.* is going to be updated to match newer major versions gradually
            result = self.ping_pong(
                connection,
                'subscription-manager list',
                expectation='Installed Product Status',
                timeout=90
            )
            if result == 'failed':
                # warn in case there's no subscription shown
                self.log[-1]['result'] = 'warning'
        else:
            self.ping_pong(
                connection,
                'subscription-manager list',
                expectation='Installed Product Status',
                timeout=90
            )
        return self.log



""" This module contains testcase_22_gpg_keys test """
from testcase import Testcase


class testcase_22_gpg_keys(Testcase):
    """
    Check that specified gpg keys are installed
    """

    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS (>=5.5, <7.0)'}
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        self.ping_pong(connection, 'grep \'^gpgcheck=\' /etc/yum.repos.d/redhat-*.repo | cut -d\= -f2 | sort -uf | tr -d \' \'', '\r\n1\r\n')
        self.ping_pong(connection, 'rpm -qa gpg-pubkey* | wc -l', '\r\n2\r\n', 10)
        self.get_return_value(connection, 'rpm -q gpg-pubkey-2fa658e0-45700c69', 30)
        if params['version'].startswith('6.'):
            self.get_return_value(connection, 'rpm -q gpg-pubkey-fd431d51-4ae0493b', 30)
        elif params['version'].startswith('5.'):
            self.get_return_value(connection, 'rpm -q gpg-pubkey-37017186-45761324', 30)
        if params['platform'].upper() == 'BETA':
            self.get_return_value(connection, 'test -f /etc/pki/rpm-gpg/RPM-GPG-KEY-redhat-beta')
        return self.log



""" This module contains testcase_02_selinux_context test """
import re
import yaml
import os
import tempfile
from testcase import Testcase
from collections import defaultdict

EXCLUDED_PATHS = defaultdict(lambda: ['/mnt', '/proc', '/sys'], {
    'ATOMIC': [ '/var/mnt', '/proc', '/sys', '/sysroot/' ]
})

def restorecon_cmd(platform):
    '''return a restorecon cmd  to execute remotely to get list of changes'''
    return 'restorecon -R -v -n ' +  ' '.join(map(lambda path: '-e ' + path, EXCLUDED_PATHS[platform])) + \
        ''' | sed -e 's, context , ,' -e 's,^restorecon reset ,,' '''

class testcase_02_selinux_context(Testcase):
    """
    Check if the kickstart restores selinux appropriately
    """
    stages = ['stage1']
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version']
        #get the restorecon output file
        self.ping_pong(connection, restorecon_cmd(prod) + " > /tmp/restorecon_output.txt && echo SUCCESS",
                       "\r\nSUCCESS\r\n", 260)
        tfile = tempfile.NamedTemporaryFile(delete=False)
        tfile.close()
        connection.sftp.get('/tmp/restorecon_output.txt', tfile.name)
        output = open(tfile.name, 'r')
        result = output.read()
        output.close()
        os.unlink(tfile.name)
        # convert output file into a dictionary to be able to compare with allowed selinux exclusions
        result = result.split("\n")
        result.pop()
        restorecon_dict = {}
        for line in result:
            filename = line.split(" ")[0]
            context = line.split(" ")[1]
            source_context = context.split(":")[2]
            destination_context = context.split(":")[5]
            restorecon_dict[filename] = [source_context, destination_context]
        #figure out if there are new/lost entries or the restorecon output matched the list of allowed exclusions
        with open(self.datadir + '/selinux_context.yaml') as selinux_context:
            # by default, no SELinux issues are expected for untracked platform/version combinations
            context_exclusions = defaultdict(lambda: {}, yaml.load(selinux_context))['%s_%s' % (prod, ver)]

        lost_entries = []
        for filename in context_exclusions:
            pattern = re.compile('%s' % filename)
            matched = False
            for filename_rest in restorecon_dict:
                match_result = pattern.match(filename_rest)
                if match_result and context_exclusions[filename] == restorecon_dict[filename_rest]:
                    matched = True
                    break
            if not matched:
                lost_entries.append([filename, context_exclusions[filename]])

        new_entries = []
        for filename_rest in restorecon_dict:
            matched = False
            for filename in context_exclusions:
                pattern = re.compile('%s' % filename)
                match_result = pattern.match(filename_rest)
                if match_result and context_exclusions[filename] == restorecon_dict[filename_rest]:
                    matched = True
                    break
            if not matched:
                new_entries.append([filename_rest, restorecon_dict[filename_rest]])

        if new_entries == [] and lost_entries != []:
            self.log.append({'result': 'warning', 'comment': '\nLost entries:' + str(lost_entries)})
        if new_entries == lost_entries == []:
            self.log.append({'result': 'passed', 'comment': 'No new or lost entries detected. Restorecon output file matches with the list of allowed SElinux context discrepancies.'})
        if new_entries != []:
            self.log.append({'result': 'failed', 'comment': '\nFail.New entries detected:' + str(new_entries) + '\nLost entries:' + str(lost_entries)})
        return self.log



""" This module contains testcase_21_disk_size_format test """
import re
from testcase import Testcase
from distutils.version import LooseVersion


def mount2fsys(prod, vers, mpoint):
    '''return expected fsys for particular mountpoint/platform/version combination'''
    if vers < '6.0':
        # ext3 for RHEL5 any mpoint
        return 'ext3'
    elif vers < '7.0' and mpoint == '/':
        # ext4 for / in RHEL6
        return 'ext4'
    elif prod in ('RHEL', 'BETA') and vers >= '7.0' and mpoint == '/':
        # xfs for / in RHEL7 and newer
        return 'xfs'
    elif prod == 'ATOMIC':
        # xfs for atomic, save for /var/mnt
        if mpoint == '/var/mnt':
            return 'ext3'
        else:
            return 'xfs'
    # ext3 for all other FS in other OSes
    return 'ext3'


class testcase_21_disk_size_format(Testcase):
    """
    Check filesystems:
    - / shoud be ext3 on RHEL5, ext4 otherwise
    - other filesystems shoud be always be ext3
    - all filesystems show have at least 4gb of available space (weird)
    """
    stages = ['stage1']
    tags = ['default']
    applicable = {'platform': '(?i)RHEL|BETA|ATOMIC'}

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        vers = LooseVersion(params['version'])

        disks = self.get_result(connection, 'mount | grep \'^/dev\' | awk \'{print $1}\'')
        if disks:
            for disk in set(disks.split()):
                # check free space
                self.get_return_value(connection, '[ `df -k %s | awk \'{ print $2 }\' | tail -n 1` -gt 3937219 ]' % disk)
                # check mount point fstype
                mpoint = self.match(connection, 'echo \'###\' ;mount | grep \'^%s\' | awk \'{print $3}\'; echo \'###\'' % disk,
                                    re.compile('.*\r\n###\r\n(.*)\r\n###\r\n.*', re.DOTALL))
                fsys = self.match(connection, 'echo \'###\' ;mount | grep \'^%s\' | awk \'{print $5}\'; echo \'###\'' % disk,
                                  re.compile('.*\r\n###\r\n(.*)\r\n###\r\n.*', re.DOTALL))
                if not mpoint or not fsys:
                    continue
                # mpoint/fsys are a match-groups tuple
                # mpoint/fsys can be multiple values on e.g. ATOMIC
                for mpoint, fsys in zip(mpoint[0].split(), fsys[0].split()):
                    expected_fsys = mount2fsys(prod, vers, mpoint)
                    self.log.append({'expected': mpoint + ': ' + expected_fsys, 'actual': fsys})
                    assert fsys == expected_fsys
                    self.log[-1]['result'] = 'passed'
        return self.log



'''
the main report function module
'''
import logging
from ..work.data import load_yaml
from ..work.common import RESULT_PASSED
from result import get_overall_result

logger = logging.getLogger(__name__)

def main(config, istream, ostream, whitelist, verbose):
    result, ami_results = get_overall_result(load_yaml(istream), whitelist, verbose)
    print >>ostream, '# overal result: %s' % result
    for ami_result, ami_log in ami_results:
        for log_line in ami_log:
            print >>ostream, log_line
    # result passed -> 0 otherwise 1;
    # please note: result == passed and 0 or 1 -> 1 always
    return result != RESULT_PASSED and 1 or 0




""" This module contains testcase_37_sshd_bug923996 test """
from testcase import Testcase
from distutils.version import LooseVersion as Version


class testcase_37_sshd_bug923996(Testcase):
    """
    Perform test against bug #923996: multiple PermitRootLogin=... in /etc/ssh/sshd_config
    """
    stages = ['stage2']
    tags = ['default']
    applicable = {
        'platform': '(?i)RHEL|BETA',
        'version': lambda ver: ver >= Version('5.5') and ver < Version('5.10') or \
                            ver >= Version('6.5') and ver < Version('7.0')
    }

    def test(self, connection, params):
        """ Perform test """

        if params['version'].startswith('5.'):
            self.get_return_value(connection, '[ `grep ^PermitRootLogin /etc/ssh/sshd_config | wc -l` -lt 2 ]')
        if params['version'].startswith('6.'):
            self.get_return_value(connection, '[ `grep ^PermitRootLogin /etc/ssh/sshd_config | wc -l` -eq 0 ]')
        return self.log



""" This module contains testcase_23_syslog test """
from testcase import Testcase


class testcase_23_syslog(Testcase):
    """
    Check /etc/rsyslog.conf checksum
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA', 'version': 'OS(>=5.5, <7.0)'}
    tags = ['default']

    def test(self, connection, params):
        """ Perform test """

        ver = params['version']
        rsyslog_md5 = self.get_result(connection, 'md5sum /etc/rsyslog.conf | cut -f 1 -d \' \'')
        rsyslog_md5_check = []
        if rsyslog_md5:
            if ver.startswith('5.'):
                rsyslog_md5_check = ['bd4e328df4b59d41979ef7202a05e074', '15936b6fe4e8fadcea87b54de495f975']
            elif ver[:3] in ['6.0', '6.1', '6.2']:
                rsyslog_md5_check = ['dd356958ca9c4e779f7fac13dde3c1b5']
            elif ver.startswith('6.'):
                rsyslog_md5_check = ['8b91b32300134e98ef4aee632ed61e21']

        if rsyslog_md5 in rsyslog_md5_check:
            self.log.append({'result': 'passed', 'comment': '/etc/rsyslog.conf md5 matches'})
        else:
            self.log.append({'result': 'failed', 'comment': '/etc/rsyslog.conf md5 differs (%s, should be %s)' % (rsyslog_md5, rsyslog_md5_check)})

        return self.log



""" This module contains testcase_11_package_set test """
import os
import re
from testcase import Testcase, SkipException
from dva.work.data import load_yaml
FILENAME_FIELDS = ['platform', 'product', 'version']


class testcase_11_package_set(Testcase):
    """
    Check that all packages specified in list
    /usr/share/dva/data/packages_<name>  are present
    """
    stages = ['stage1']
    applicable = {'platform': '(?i)RHEL|BETA|FEDORA'}
    tags = ['default']

    def package_list_filename(self, params):
        '''construct package list filename based on the params provided'''
        filename = 'packages_' + '_'.join([params[field].lower() for field in FILENAME_FIELDS]) + '.yaml'
        # beta packages are the same as rhel
        filename = filename.replace('beta', 'rhel')
        path = os.path.sep.join([self.datadir, filename])
        if not os.path.exists(path):
            raise SkipException('packages list %s not found' % path)
        return path

    def required_packages(self, params):
        '''return set of required packages'''
        with open(self.package_list_filename(params)) as fd:
            ret = set(load_yaml(fd))
        return ret
        

    def test(self, connection, params):
        """ Perform test """
        # load expected package list
        package_set_required = self.required_packages(params)

        # fetch packages installed on the instance
        packages_log = self.match(connection,
                              'rpm -qa --queryformat \'%{NAME},\' && echo',
                              re.compile('.*\r\n(.*),\r\n.*', re.DOTALL),
                              timeout=30)[0]
        package_set_got = set(packages_log.split(','))

        # augment userdata-installed packages
        if params['userdata'].find('yum -y install xdelta') != -1:
            package_set_got.discard('xdelta')

        difference = package_set_required.difference(package_set_got)
        difference_new = package_set_got.difference(package_set_required)
        
        self.log.append({'result': 'passed', 'comment': 'Newly introduced packages: ' + str(sorted(list(difference_new)))})
        if params['platform'].upper() == 'BETA' and len(difference) > 1:
            self.log.append({'result': 'failed', 'comment': 'Beta may lack not more than 1 package: ' + str(sorted(list(difference)))})
        elif len(difference) > 1:
            self.log.append({'result': 'failed', 'comment': 'some packages are missing: ' + str(sorted(list(difference)))})
        else:
            self.log.append({'result': 'passed', 'comment': 'All required package are included'})
        return self.log



""" This module contains testcase_361_ebs_deferred_detach test """
import time
import multiprocessing
import logging
from testcase import Testcase
from dva.work.common import RESULT_FAILED


class testcase_361_ebs_deferred_detach(Testcase):
    """
    Perform EBS test:
    - Create the volume
    - Attach the volume
    - Use the volume
    - Try to detach the volume
    - Try to umount the volume
    - Remove volume
    """
    stages = ['stage1']
    tags = ['kernel']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version'].upper()
        device = '/dev/sdk'
        if params['cloudhwname'] == 'hs1.8xlarge':
            for dev in params['bmap']:
                if dev['name'] == device:
                    device = '/dev/xvdac'
                    break
        ec2connection = params['instance']['connection']
        if 'placement' in params['instance']:
            volume = ec2connection.create_volume(3, params['instance']['placement'])
        elif '_placement' in params['instance']:
            volume = ec2connection.create_volume(3, params['instance']['_placement'])
        else:
            self.log.append({'result': RESULT_FAILED,
                             'comment': 'Failed to get instance placement'
                             })
            return self.log
        self.logger.debug(multiprocessing.current_process().name + ': Volume %s created' % volume.id)
        time.sleep(5)
        volume.update()
        wait = 0
        while volume.volume_state() == 'creating':
            time.sleep(1)
            wait += 1
            if wait > 300:
                self.log.append({'result': RESULT_FAILED,
                                 'comment': 'Failed to create EBS volume %s (timeout 300)' % volume.id
                                 })
                ec2connection.delete_volume(volume.id)
                return self.log
        if volume.volume_state() == 'available':
            self.logger.debug(multiprocessing.current_process().name + ': Ready to attach %s: %s %s' % (volume.id,
                                                                                                    volume.volume_state(),
                                                                                                    volume.attachment_state()))
            ec2connection.attach_volume(volume.id, params['instance']['id'], device)
            time.sleep(5)
            volume.update()
            wait = 0
            while volume.attachment_state() == 'attaching':
                volume.update()
                self.logger.debug(multiprocessing.current_process().name + ': Wait attaching %s: %s %s' % (volume.id,
                                                                                                       volume.volume_state(),
                                                                                                       volume.attachment_state()))
                time.sleep(1)
                wait += 1
                if wait > 300:
                    self.log.append({'result': 'failure',
                                     'comment': 'Failed to attach EBS volume %s (timeout 300)' % volume.id
                                     })
                    ec2connection.delete_volume(volume.id)
                    return self.log
            if volume.attachment_state() != 'attached':
                self.logger.debug(multiprocessing.current_process().name + ': Error attaching volume %s' % volume.id)
                self.log.append({'result': 'failure',
                                 'comment': 'Failed to attach EBS volume %s' % volume.id
                                 })
                ec2connection.delete_volume(volume.id)
                return self.log

            if (prod in ['RHEL', 'BETA']) and (ver.startswith('5.')):
                name = device
            elif (prod in ['RHEL', 'BETA']) and (ver.startswith('6.')) and (params['virtualization'] != 'hvm'):
                # 4-letter shift
                name = device.replace("/dev/sd", "/dev/xvd")[:-1] + chr(ord(device.replace("/dev/sd", "/dev/xvd")[-1:]) + 4)
            else:
                name = device.replace("/dev/sd", "/dev/xvd")
            # waiting for this volume
            for _ in xrange(20):
                if self.get_return_value(connection, 'ls -l %s' % name, 30, nolog=True) == 0:
                    break
                time.sleep(1)
            self.get_return_value(connection, 'ls -l %s' % name, 30)
            self.get_return_value(connection, 'echo y| mkfs.ext3 %s' % name, 300)

            self.get_return_value(connection, 'mkdir /mnt/deferred')
            self.get_return_value(connection, 'mount %s /mnt/deferred' % name)
            self.get_return_value(connection, 'dd if=/dev/zero of=/mnt/deferred/zzz_file bs=512 count=102400')

            self.logger.debug(multiprocessing.current_process().name + ': Ready to detach %s: %s %s' % (volume.id,
                                                                                                    volume.volume_state(),
                                                                                                    volume.attachment_state()))
            ec2connection.detach_volume(volume.id)
            time.sleep(20)
            volume.update()

            if self.ping_pong(connection, 'umount /mnt/deferred && echo SUCCESS', '\r\nSUCCESS\r\n', 20) is None:
                # bug#794803
                # doing force-detach
                ec2connection.detach_volume(volume.id, force=True)

            wait = 0
            while volume.attachment_state() == 'detaching':
                volume.update()
                self.logger.debug(multiprocessing.current_process().name + ': Wait detaching %s: %s %s' % (volume.id,
                                                                                                       volume.volume_state(),
                                                                                                       volume.attachment_state()))
                time.sleep(1)
                wait += 1
                if wait > 300:
                    self.log.append({'result': 'failure',
                                     'comment': 'Failed to detach EBS volume %s (timeout 300)' % volume.id
                                     })
                    ec2connection.delete_volume(volume.id)
                    return self.log
            if volume.volume_state() != 'available':
                self.logger.debug(multiprocessing.current_process().name + ': Error detaching volume %s' % volume.id)
                self.log.append({'result': 'failure',
                                 'comment': 'Failed to detach EBS volume %s' % volume.id
                                 })
                return self.log
            self.logger.debug(multiprocessing.current_process().name + ': Ready to delete %s: %s %s' % (volume.id,
                                                                                                    volume.volume_state(),
                                                                                                    volume.attachment_state()))
            if not ec2connection.delete_volume(volume.id):
                self.log.append({'result': 'failure',
                                 'comment': 'Failed to remove EBS volume %s' % volume.id
                                 })
        else:
            self.log.append({'result': 'failure',
                             'comment': 'Failed to create EBS volume %s' % volume.id
                             })
        return self.log



""" This module contains testcase_32_ephemeral test """
from testcase import Testcase
from os.path import basename


class testcase_32_ephemeral(Testcase):
    """
    It should be possible to use ephemeral device (if we have one)
    Note that in rhel6.5 there is no shift letter in dick device name
    """
    stages = ['stage1']
    tags = ['default']
    after = ['testcase_21_disk_size_format']

    def test(self, connection, params):
        """ Perform test """

        prod = params['platform'].upper()
        ver = params['version'].upper()
        ephemerals = []

        if 'bmap' in params:
            ephemerals = [bdev for bdev in params['bmap'] if 'ephemeral_name' in bdev]

        if not ephemerals:
            # are there are some devices to check?
            self.log.append({'result': 'skip',
                             'comment': 'no ephemeral devices in block map'
                             })
            return self.log

        # figure out what fstype to use with focus on mkfs.<fstype> speed
        fstype = None
        for fstype in ['vfat', 'xfs', 'ext3']:
            if self.get_result(connection, 'ls -la /sbin/mkfs.%s 2> /dev/null | wc -l' % fstype, 5) == '1':
                break

        if fstype == 'vfat':
            # so that mkfs.vfat /dev/<device> doesn't complain
            # because of making fs on whole drive instead of a partition
            mkfs_opt = '-I'
        else:
            mkfs_opt = ''

        devices = []
        for bdev in ephemerals:
            name = bdev['name']
            if (prod in ['RHEL', 'BETA']) and (ver.startswith('5.')):
                if name.startswith('/dev/xvd'):
                    # no xvd* for RHEL5
                    continue
            elif (prod in ['RHEL', 'BETA']) and (ver.startswith('6.') and ver != '6.0'):
                if name.startswith('/dev/sd'):
                    name = '/dev/xvd' + name[7:]
                if params['virtualization'] != 'hvm' and ver != '6.5' and len(name) == 9 and ord(name[8]) < ord('w'):
                    # there is a 4-letter shift
                    name = name[:8] + chr(ord(name[8]) + 4)
            else:
                # Fedora and newer RHELs
                if name.startswith('/dev/sd'):
                    name = '/dev/xvd' + name[7:]
            # test: check device presence
            self.get_return_value(connection, 'fdisk -l %s | grep \'^Disk\'' % name, 30)
            if self.get_result(connection, 'grep \'%s \' /proc/mounts  | wc -l' % name, 5) == '0':
                # device is not mounted, doing fs creation
                devices.append(name)

        if not devices:
            return self.log

        mkfs_commands = ['mkfs.%s %s %s' % (fstype, mkfs_opt, name) for name in devices]
        assert self.ping_pong(connection, ' & '.join(mkfs_commands) + ' & echo MKFS', '(?s).*\r\nMKFS.*'), "call mkfs_commands failed"
        assert self.ping_pong(connection, 'wait && echo WAIT', '(?s).*\r\nWAIT.*', 1000), "wait failed?!?!"
        dest_names = ['/tmp/mnt-%s' % basename(name) for name in devices]
        self.get_return_value(connection, 'mkdir -p ' + ' '.join(dest_names))
        mount_commands = ['mount -t %s %s /tmp/mnt-%s' % (fstype, name, basename(name)) for name in devices]
        self.get_return_value(connection, ' && '.join(mount_commands))
        return self.log



""" All testing modules """

import testcase_01_bash_history
import testcase_02_selinux_context
import testcase_03_running_services
import testcase_04_cloud_firstboot
import testcase_05_grub
import testcase_06_inittab
import testcase_07_libc6_xen_conf
import testcase_08_memory
import testcase_09_nameserver
import testcase_10_networking
import testcase_11_package_set
import testcase_12_passwd_group
import testcase_13_resize2fs
import testcase_14_host_details
import testcase_15_rhel_version
import testcase_16_selinux
import testcase_17_shells
import testcase_18_sshd
import testcase_19_rhn_system_id
import testcase_20_auditd
import testcase_21_disk_size_format
import testcase_22_gpg_keys
import testcase_23_syslog
import testcase_24_yum_plugin
import testcase_25_uname
import testcase_26_verify_rpms
import testcase_27_yum_repos
import testcase_28_iptables
import testcase_29_swap_file
import testcase_30_rhn_certificates
import testcase_31_subscription_management
import testcase_32_ephemeral
import testcase_33_userdata
import testcase_34_cpu
import testcase_35_console
import testcase_360_ebs
import testcase_361_ebs_deferred_detach
import testcase_37_sshd_bug923996
import testcase_38_rclocal_checksum
import testcase_39_root_is_locked
import testcase_41_rh_amazon_rhui_client
import testcase_42_ipv6
import testcase_43_yum_versionlock
import testcase_50_yum_package_install
import testcase_55_yum_group_install
import testcase_60_yum_update
import testcase_61_yum_proxy
import testcase_62_cpuflags
import testcase_80_no_avc_denials
import testcase_90_1_kernel_bug874053
import testcase_90_2_cgroups_kernel_bug914737
import testcase_90_3_bug1074660
import testcase_97_collect_packagelist
import testcase_98_dmesg_stage1
import testcase_98_kernel_upgrade_pre
import testcase_98_package_install_pre
import testcase_98_kernel_parameter



import logging
logger = logging.getLogger(__name__)

def profiled(fn):
    def profiled_wrapper(*args, **kvs):
        import cProfile
        import pstats
        import os
        import tempfile
        profiler = cProfile.Profile()
        profiler.enable()
        try:
            ret = fn(*args, **kvs)
        finally:
            profiler.disable()
            # for multiprocessing tracking, each process should have a separate file
            _, name = tempfile.mkstemp(prefix="profiler-%s-" % os.getpid(), suffix=".log")
            profiler.dump_stats(name)
        return ret
    return profiled_wrapper


def memprofiledCls(cls):
    '''
    memory profile a class
    dumps stats atexit
    '''
    def profiled_wrapper(*args, **kvs):
        '''
        wraps original class-definition callable
        '''
        from pympler.classtracker import ClassTracker
        import atexit

        tracker = ClassTracker()
        tracker.track_class(cls)

        # monkeypatch the original class __setattr__
        # to take mem usage snapshots at setting attributes
        original_setattr = cls.__setattr__
        def __setattr__(self, attrname, value):
            '''
            snapshot mem usage at attrname access
            '''
            try:
                tracker.create_snapshot(attrname)
            except Exception as e:
                pass
            return original_setattr(self, attrname, value)

        cls.__setattr__ = __setattr__

        def dump_stats_atexit():
            '''
            dumps mem stats at program exit for each process separately
            '''
            import os
            import tempfile
            _, name = tempfile.mkstemp(prefix="mem-%s-%s-" % (cls.__name__, os.getpid()), suffix=".log")
            tracker.stats.dump_stats(name)

        # register the stats dumping
        atexit.register(dump_stats_atexit)
        # back to class instantiation
        return cls(*args, **kvs)
    return profiled_wrapper





'''
cache and tools
'''
from gevent import getcurrent, socket

import sys
import logging
import traceback
import paramiko
from ..tools.retrying import retrying, EAgain
from .. import cloud, work
from stitches import Connection, Expect, ExpectFailed
from stitches.connection import StitchesConnectionException
from weakref import WeakValueDictionary


CONNECTION_ATTEMPTS = 120
CONNECTION_ATTEMPTS_RETRY_AFTER = 1
CONNECTION_CACHE = WeakValueDictionary()

logger = logging.getLogger(__name__)

class ConnectionCacheError(Exception):
    '''bad things happen in connection cache'''

def connection_cache_key(params):
    '''params to connection cache key tuple: (None, host, user, key)'''
    try:
        return id(getcurrent()), params['hostname'], params['ssh']['user'], params['ssh']['keyfile']
    except KeyError as err:
        # some key is missing in params
        raise ConnectionCacheError('params missing key: %s' % err)

def drop_connection(params):
    """ Close&remove connection """
    key = connection_cache_key(params)
    try:
        CONNECTION_CACHE.pop(key).disconnect()
        logger.debug('dropped: %s', key)
    except KeyError as err:
        pass

def assert_connection(connection):
    '''assert a connection is alive; wel... ;) at a point in time'''
    try:
        logger.debug('asserting connection: %s', connection)
        Expect.ping_pong(connection, 'uname', 'Linux')
        logger.debug('asserting connection: %s passed', connection)
    except (IOError, socket.error, socket.timeout, StitchesConnectionException, ExpectFailed, \
            paramiko.ssh_exception.AuthenticationException, paramiko.ssh_exception.SSHException, EOFError) as err:
        logger.debug('asserting %s got: %s(%s): %s', connection, type(err), err, traceback.format_exc())
        raise EAgain(err)

@retrying(maxtries=CONNECTION_ATTEMPTS, sleep=CONNECTION_ATTEMPTS_RETRY_AFTER, loglevel=logging.DEBUG,
            final_exception=ConnectionCacheError)
def get_connection(params):
    '''Get connection connection based on params. Retrying.'''
    # maybe the connection is cached...
    key = connection_cache_key(params)
    try:
        connection = CONNECTION_CACHE[key]
        logger.debug('cache hit: %s', key)
    except KeyError as err:
        logger.debug('cache miss: %s', key)
        _, host, user, ssh_key = key
        CONNECTION_CACHE[key] = connection =  Connection(host, user, ssh_key)
        logger.debug('got new connection: %s', connection)

    try:
        # make sure the connection is alive
        assert_connection(connection)
        logger.debug('got alive connection: %s', connection)
    except EAgain as err:
        # something went wrong --- drop and keep retrying
        logger.debug('got %s asserting %s --- dropping connection', err, connection)
        drop_connection(params)
        work.params.reload(params)
        raise err
    # got connection --- return
    return connection
